__author__ = "David Rusk <drusk@uvic.ca>"

import os
import unittest

from hamcrest import (assert_that, is_in, is_not, equal_to, is_, none,
                      contains_inanyorder, has_length, contains)
from mock import Mock, call, MagicMock

from tests.base_tests import FileReadingTestCase, DirectoryCleaningTestCase
from tests.testutil import CopyingMock
from ossos.gui import tasks
from ossos.gui.context import WorkingContext, LocalDirectoryWorkingContext
from ossos.gui.models.collections import StatefulCollection
from ossos.gui.models.exceptions import NoAvailableWorkException
from ossos.gui.models.validation import ValidationModel
from ossos.astrom import (AstromParser, StreamingAstromWriter, SourceReading,
                          Source)
from ossos.mpc import MPCWriter
from ossos.gui.models.imagemanager import ImageManager
from ossos.gui.progress import LocalProgressManager, InMemoryProgressManager
from ossos.gui.models.workload import (WorkUnitProvider, WorkUnit,
                                       RealsWorkUnit, CandidatesWorkUnit,
                                       RealsWorkUnitBuilder,
                                       PreFetchingWorkUnitProvider, TracksWorkUnitBuilder)
from ossos.ssos import SSOSData, SSOSParser


class TestDirectoryManager(object):
    def __init__(self):
        self.listings = {}

    def set_listing(self, suffix, listing):
        self.listings[suffix] = listing[:]

    def get_listing(self, suffix):
        return self.listings[suffix][:]

    def get_full_path(self, filename):
        return filename

    def get_file_size(self, filename):
        return 1


class TestWorkUnitBuilder(object):
    def build_workunit(self, full_path):
        _, filename = os.path.split(full_path)
        workunit = Mock(spec=WorkUnit)
        workunit.get_filename.return_value = filename
        return workunit


class AbstractWorkUnitTest(FileReadingTestCase):
    def setUp(self):
        self.testfile = self.get_input_file()
        parser = AstromParser()
        self.data = parser.parse(self.get_abs_path(self.testfile))
        self.progress_manager = Mock(spec=LocalProgressManager)
        self.progress_manager.get_processed_indices.return_value = []
        self.output_context = Mock(spec=WorkingContext)

    def get_input_file(self):
        raise NotImplementedError()


class WorkUnitTest(AbstractWorkUnitTest):
    def get_input_file(self):
        return "data/1584431p15.measure3.reals.astrom"

    def setUp(self):
        super(WorkUnitTest, self).setUp()

        self.workunit = RealsWorkUnit(self.testfile, self.data,
                                      self.progress_manager,
                                      self.output_context)
        self.writer = Mock(spec=MPCWriter)
        self.workunit.get_writer = Mock(return_value=self.writer)

    def test_initialization(self):
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_source_count(), equal_to(3))
        assert_that(self.workunit.get_obs_count(), equal_to(3))

    def test_next_source_previous_source(self):
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        self.workunit.next_source()
        assert_that(self.workunit.get_current_source_number(), equal_to(1))
        self.workunit.previous_source()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))

    def test_next_source_wrap(self):
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        self.workunit.next_source()
        assert_that(self.workunit.get_current_source_number(), equal_to(1))
        self.workunit.next_source()
        assert_that(self.workunit.get_current_source_number(), equal_to(2))
        self.workunit.next_source()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))

    def test_next_obs(self):
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))
        self.workunit.next_obs()
        assert_that(self.workunit.get_current_obs_number(), equal_to(1))
        self.workunit.next_obs()
        assert_that(self.workunit.get_current_obs_number(), equal_to(2))
        self.workunit.next_obs()
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

    def test_previous_source_wrap(self):
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        self.workunit.previous_source()
        assert_that(self.workunit.get_current_source_number(), equal_to(2))
        self.workunit.previous_source()
        assert_that(self.workunit.get_current_source_number(), equal_to(1))
        self.workunit.previous_source()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))

    def test_previous_obs(self):
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))
        self.workunit.previous_obs()
        assert_that(self.workunit.get_current_obs_number(), equal_to(2))
        self.workunit.previous_obs()
        assert_that(self.workunit.get_current_obs_number(), equal_to(1))
        self.workunit.previous_obs()
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

    def test_next_source_resets_obs(self):
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))
        self.workunit.next_obs()
        assert_that(self.workunit.get_current_obs_number(), equal_to(1))
        self.workunit.next_source()
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))
        self.workunit.next_obs()
        assert_that(self.workunit.get_current_obs_number(), equal_to(1))


class RealsWorkUnitTest(AbstractWorkUnitTest):
    def get_input_file(self):
        return "data/1584431p15.measure3.reals.astrom"

    def setUp(self):
        super(RealsWorkUnitTest, self).setUp()

        self.workunit = RealsWorkUnit(self.testfile, self.data,
                                      self.progress_manager,
                                      self.output_context)
        self.writer = Mock(spec=MPCWriter)
        self.workunit.get_writer = Mock(return_value=self.writer)

    def test_next_vettable_item_no_validation(self):
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

        self.workunit.next_item()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(1))

        self.workunit.next_item()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(2))

        self.workunit.next_item()
        # Should have looped back to first observation of the same source
        # because we haven't finished processing it.
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

    def test_next_vettable_item_after_validate_last(self):
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

        self.workunit.next_item()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(1))

        self.workunit.next_item()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(2))

        self.workunit.accept_current_item()
        self.workunit.next_item()

        # Should have looped back to first observation of the same source
        # because we haven't finished processing it.
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

        self.workunit.accept_current_item()
        self.workunit.next_item()

        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(1))

        self.workunit.accept_current_item()
        self.workunit.next_item()

        # We already validated the last reading, so we should be jumping
        # straight to the second source now.

        assert_that(self.workunit.get_current_source_number(), equal_to(1))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

        self.workunit.next_item()

        assert_that(self.workunit.get_current_source_number(), equal_to(1))
        assert_that(self.workunit.get_current_obs_number(), equal_to(1))

    def test_next_vettable_item_jump_over_processed(self):
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

        self.workunit.next_item()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(1))

        self.workunit.reject_current_item()
        self.workunit.next_item()

        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(2))

        self.workunit.next_item()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

        self.workunit.next_item()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(2))

    def test_accept_current_item(self):
        first_source = self.data.get_sources()[0]
        first_item = first_source.get_reading(0)
        second_item = first_source.get_reading(1)

        assert_that(self.workunit.is_item_processed(first_item), equal_to(False))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(False))

        self.workunit.accept_current_item()

        assert_that(self.workunit.is_item_processed(first_item), equal_to(True))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(False))

        self.workunit.next_item()
        self.workunit.accept_current_item()

        assert_that(self.workunit.is_item_processed(first_item), equal_to(True))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(True))

    def test_reject_current_item(self):
        first_source = self.data.get_sources()[0]
        first_item = first_source.get_reading(0)
        second_item = first_source.get_reading(1)

        assert_that(self.workunit.is_item_processed(first_item), equal_to(False))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(False))

        self.workunit.reject_current_item()

        assert_that(self.workunit.is_item_processed(first_item), equal_to(True))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(False))

        self.workunit.next_item()
        self.workunit.reject_current_item()

        assert_that(self.workunit.is_item_processed(first_item), equal_to(True))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(True))

    def test_get_current_item_index(self):
        for index in range(8):
            assert_that(self.workunit.get_current_item_index(), equal_to(index))
            self.workunit.accept_current_item()
            self.workunit.next_item()

        assert_that(self.workunit.get_current_item_index(), equal_to(8))

    def test_get_unprocessed_sources(self):
        sources = self.data.get_sources()

        assert_that(self.workunit.get_unprocessed_sources(),
                    contains_inanyorder(*sources))

        self.workunit.accept_current_item()
        assert_that(self.workunit.get_unprocessed_sources(),
                    contains_inanyorder(*sources))

        self.workunit.next_item()
        self.workunit.accept_current_item()
        self.workunit.next_item()
        self.workunit.accept_current_item()

        assert_that(self.workunit.get_unprocessed_sources(),
                    contains_inanyorder(sources[1], sources[2]))

        self.workunit.next_item()
        self.workunit.reject_current_item()
        self.workunit.next_item()
        self.workunit.accept_current_item()
        self.workunit.next_item()
        self.workunit.accept_current_item()

        assert_that(self.workunit.get_unprocessed_sources(),
                    contains_inanyorder(sources[2]))

        self.workunit.next_item()
        self.workunit.accept_current_item()
        self.workunit.next_item()
        self.workunit.accept_current_item()
        self.workunit.next_item()
        self.workunit.accept_current_item()

        assert_that(self.workunit.get_unprocessed_sources(),
                    has_length(0))

    def test_finish_workunit_unlocks_file(self):
        num_items = 9
        while num_items > 1:
            self.workunit.accept_current_item()
            self.workunit.next_item()
            num_items -= 1

        assert_that(self.progress_manager.unlock.called, equal_to(False))

        self.workunit.accept_current_item()

        self.progress_manager.unlock.assert_called_once_with(
            self.testfile, async=True)

    def test_flush_only_after_source_finished(self):
        assert_that(self.writer.flush.called, equal_to(False))

        self.workunit.accept_current_item()
        self.workunit.next_item()

        assert_that(self.writer.flush.called, equal_to(False))

        self.workunit.reject_current_item()
        self.workunit.next_item()

        assert_that(self.writer.flush.called, equal_to(False))

        self.workunit.accept_current_item()
        assert_that(self.writer.flush.call_count, equal_to(1))

    def test_get_output_filename_dry_run(self):
        self.workunit = RealsWorkUnit(self.testfile, self.data,
                                      self.progress_manager,
                                      self.output_context,
                                      dry_run=True)
        source = self.data.get_sources()[0]
        provisional_name = "ABCD123"
        source.set_provisional_name(provisional_name)

        assert_that(self.workunit.get_output_filename(source),
                    equal_to("1584431p15.measure3.reals.astrom.ABCD123.mpc"))

    def test_get_output_filename_not_dry_run(self):
        self.workunit = RealsWorkUnit(self.testfile, self.data,
                                      self.progress_manager,
                                      self.output_context,
                                      dry_run=False)
        source = self.data.get_sources()[0]
        provisional_name = "ABCD123"
        source.set_provisional_name(provisional_name)

        assert_that(self.workunit.get_output_filename(source),
                    equal_to("1584431p15.measure3.reals.astrom.ABCD123.mpc"))


class CandidatesWorkUnitTest(AbstractWorkUnitTest):
    def get_input_file(self):
        return "data/1584431p15.measure3.cands.astrom"

    def setUp(self):
        super(CandidatesWorkUnitTest, self).setUp()

        self.workunit = CandidatesWorkUnit(self.testfile, self.data,
                                           self.progress_manager,
                                           self.output_context)
        self.writer = Mock(spec=StreamingAstromWriter)
        self.workunit.get_writer = Mock(return_value=self.writer)

    def test_output_filename(self):
        assert_that(self.workunit.get_output_filename(),
                    equal_to(self.testfile.replace(".cands", ".reals")))

    def test_next_vettable_item(self):
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

        self.workunit.next_item()
        assert_that(self.workunit.get_current_source_number(), equal_to(1))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

        self.workunit.next_item()
        assert_that(self.workunit.get_current_source_number(), equal_to(2))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

        self.workunit.next_item()
        assert_that(self.workunit.get_current_source_number(), equal_to(0))
        assert_that(self.workunit.get_current_obs_number(), equal_to(0))

    def test_accept_current_item(self):
        first_item = self.data.get_sources()[0]
        second_item = self.data.get_sources()[1]

        assert_that(self.workunit.is_item_processed(first_item), equal_to(False))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(False))

        self.workunit.accept_current_item()

        assert_that(self.workunit.is_item_processed(first_item), equal_to(True))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(False))

        self.workunit.next_item()
        self.workunit.accept_current_item()

        assert_that(self.workunit.is_item_processed(first_item), equal_to(True))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(True))

    def test_reject_current_item(self):
        first_item = self.data.get_sources()[0]
        second_item = self.data.get_sources()[1]

        assert_that(self.workunit.is_item_processed(first_item), equal_to(False))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(False))

        self.workunit.reject_current_item()

        assert_that(self.workunit.is_item_processed(first_item), equal_to(True))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(False))

        self.workunit.next_item()
        self.workunit.reject_current_item()

        assert_that(self.workunit.is_item_processed(first_item), equal_to(True))
        assert_that(self.workunit.is_item_processed(second_item), equal_to(True))

    def test_get_unprocessed_sources(self):
        sources = self.data.get_sources()

        assert_that(self.workunit.get_unprocessed_sources(),
                    contains_inanyorder(*sources))

        self.workunit.accept_current_item()

        assert_that(self.workunit.get_unprocessed_sources(),
                    contains_inanyorder(sources[1], sources[2]))

        self.workunit.next_item()
        self.workunit.reject_current_item()

        assert_that(self.workunit.get_unprocessed_sources(),
                    contains_inanyorder(sources[2]))

        self.workunit.next_item()
        self.workunit.accept_current_item()

        assert_that(self.workunit.get_unprocessed_sources(),
                    has_length(0))

    def test_finish_workunit_unlocks_file(self):
        self.workunit.accept_current_item()
        self.workunit.next_item()
        self.workunit.accept_current_item()
        self.workunit.next_item()
        self.workunit.accept_current_item()

        self.progress_manager.unlock.assert_called_once_with(
            self.testfile, async=True)


class StatefulCollectionTest(unittest.TestCase):
    def test_basics(self):
        items = [1, 2, 3]
        undertest = StatefulCollection(items)

        assert_that(undertest[0], equal_to(items[0]))
        assert_that(undertest[1], equal_to(items[1]))
        assert_that(undertest[2], equal_to(items[2]))

        assert_that(undertest, has_length(3))

        assert_that(undertest.get_index(), equal_to(0))
        assert_that(undertest.get_current_item(), equal_to(items[0]))

        undertest.next()
        assert_that(undertest.get_index(), equal_to(1))
        assert_that(undertest.get_current_item(), equal_to(items[1]))

        undertest.previous()
        assert_that(undertest.get_index(), equal_to(0))
        assert_that(undertest.get_current_item(), equal_to(items[0]))

    def test_start_empty(self):
        undertest = StatefulCollection()

        assert_that(undertest, has_length(0))

        assert_that(undertest.get_current_item(), is_(none()))
        assert_that(undertest.get_index(), equal_to(-1))

        item1 = 1

        undertest.append(item1)
        assert_that(undertest.get_current_item(), equal_to(item1))
        assert_that(undertest.get_index(), equal_to(0))

        item2 = 2
        undertest.append(item2)

        undertest.next()
        assert_that(undertest.get_index(), equal_to(1))
        assert_that(undertest.get_current_item(), equal_to(item2))


class WorkloadManagementTest(unittest.TestCase):
    def setUp(self):
        self.progress_manager = InMemoryProgressManager(Mock(spec=LocalDirectoryWorkingContext))
        self.workunit_provider = Mock(spec=WorkUnitProvider)

        self.workunit1 = MagicMock(spec=WorkUnit)
        self.file1 = "file1"
        self.workunit1.get_filename.return_value = self.file1

        self.workunit2 = MagicMock(spec=WorkUnit)
        self.file2 = "file2"
        self.workunit2.get_filename.return_value = self.file2

        workunits = [self.workunit1, self.workunit2]

        def get_workunit(index):
            workunit = workunits[index]
            self.progress_manager.lock(workunit.get_filename())
            return workunit

        self.workunit_provider.get_workunit.side_effect = (get_workunit(index) for index in xrange(2))
        image_manager = Mock(spec=ImageManager)

        self.undertest = ValidationModel(self.workunit_provider, image_manager, None)
        self.undertest.start_work()

    def test_workunits_on_demand(self):
        assert_that(self.undertest.get_current_workunit(), equal_to(self.workunit1))
        assert_that(self.workunit_provider.get_workunit.call_count, equal_to(1))

        self.undertest.next_workunit()
        assert_that(self.undertest.get_current_workunit(), equal_to(self.workunit2))
        assert_that(self.workunit_provider.get_workunit.call_count, equal_to(2))

    def test_shift_locks(self):
        assert_that(self.undertest.get_current_workunit(), equal_to(self.workunit1))
        assert_that(self.progress_manager.owns_lock(self.file1), equal_to(True))
        assert_that(self.progress_manager.owns_lock(self.file2), equal_to(False))

        self.undertest.next_workunit()
        assert_that(self.undertest.get_current_workunit(), equal_to(self.workunit2))

        # Note we don't give up lock just by going to next workunit.  It is
        # released once it is done processing.
        assert_that(self.progress_manager.owns_lock(self.file1), equal_to(True))
        assert_that(self.progress_manager.owns_lock(self.file2), equal_to(True))


class WorkUnitProviderTest(unittest.TestCase):
    def setUp(self):
        self.taskid = "id"
        self.file1 = "file1"
        self.file2 = "file2"
        self.file3 = "file3"
        self.file4 = "file4"
        self.test_files = [self.file1, self.file2]

        directory_manager = TestDirectoryManager()
        progress_manager = InMemoryProgressManager(directory_manager)
        builder = TestWorkUnitBuilder()

        self.undertest = WorkUnitProvider(self.taskid, directory_manager,
                                          progress_manager, builder)
        self.directory_manager = directory_manager
        self.progress_manager = progress_manager
        self.directory_manager.set_listing(self.taskid, self.test_files)

    def test_create_workload_acquires_lock(self):
        self.directory_manager.set_listing(self.taskid, self.test_files)
        workunit1 = self.undertest.get_workunit()
        assert_that(self.progress_manager.owns_lock(workunit1.get_filename()),
                    equal_to(True))

    def test_create_workload_fresh_directory(self):
        workunit1 = self.undertest.get_workunit()
        assert_that(workunit1.get_filename(), is_in(self.test_files))
        self.progress_manager.record_done(workunit1.get_filename())

        workunit2 = self.undertest.get_workunit()
        assert_that(workunit2.get_filename(), is_in(self.test_files))
        assert_that(workunit2.get_filename(),
                    is_not(equal_to(workunit1.get_filename())))
        self.progress_manager.record_done(workunit2.get_filename())

        self.assertRaises(NoAvailableWorkException, self.undertest.get_workunit)

    def test_create_workload_one_file_already_done(self):
        self.progress_manager.done.add(self.file1)

        workunit = self.undertest.get_workunit()
        assert_that(workunit.get_filename(), equal_to(self.file2))
        self.progress_manager.record_done(self.file2)

        self.assertRaises(NoAvailableWorkException, self.undertest.get_workunit)

    def test_create_workload_locked_files(self):
        self.progress_manager.add_external_lock(self.file2)

        workunit = self.undertest.get_workunit()
        assert_that(workunit.get_filename(), equal_to(self.file1))
        self.progress_manager.record_done(self.file1)

        self.assertRaises(NoAvailableWorkException, self.undertest.get_workunit)

    def test_items_in_ignore_list_ignored(self):
        workunit = self.undertest.get_workunit(ignore_list=[self.file1])

        assert_that(workunit.get_filename(), equal_to(self.file2))
        self.progress_manager.record_done(self.file2)

        self.assertRaises(NoAvailableWorkException, self.undertest.get_workunit,
                          ignore_list=[self.file1])

    def test_file_found_to_be_done_not_checked_again(self):
        test_files = [self.file1, self.file2, self.file3, self.file4]
        self.directory_manager.set_listing(self.taskid, test_files)

        self.progress_manager.lock(self.file2)
        self.progress_manager.record_done(self.file2)
        self.progress_manager.unlock(self.file2)

        # We don't yet know file1 is done.
        assert_that(self.undertest.get_potential_files([]),
                    contains_inanyorder(self.file1, self.file2, self.file3,
                                        self.file4))
        assert_that(self.undertest.get_workunit().get_filename(), equal_to(self.file1))

        # We have not yet discovered file2 is done because we found file 1
        # right away.  However, we should remember we already returned file1.
        assert_that(self.undertest.get_potential_files([]),
                    contains_inanyorder(self.file2, self.file3, self.file4))

        # Here we should discover file2 is done and skip over it
        assert_that(self.undertest.get_workunit().get_filename(), equal_to(self.file3))

        # So the next time we know not to check file2 again
        assert_that(self.undertest.get_potential_files([]),
                    contains_inanyorder(self.file4))


class PreFetchingWorkUnitProviderTest(unittest.TestCase):
    def setUp(self):
        self.prefetch_quantity = 2
        self.workunit_provider = CopyingMock(spec=WorkUnitProvider)
        self.undertest = PreFetchingWorkUnitProvider(self.workunit_provider,
                                                     self.prefetch_quantity)
        self._workunit_number = 0

    def create_workunit(self, num=None):
        if num is None:
            num = self._workunit_number

        workunit = Mock(spec=WorkUnit)
        workunit.get_filename.return_value = "Workunit%d" % num
        self._workunit_number += 1
        return workunit

    def set_workunit_provider_return_values(self, return_values):
        def side_effect(*args, **kwargs):
            result = return_values.pop(0)
            if isinstance(result, Exception):
                raise result
            return result

        self.workunit_provider.get_workunit.side_effect = side_effect

    def mock_prefetch_workunit(self, bypass_threading=False):
        prefetch_workunit_mock = Mock()

        if bypass_threading:
            def do_prefetch():
                self.undertest._do_prefetch_workunit()

            prefetch_workunit_mock.side_effect = do_prefetch

        else:
            def generate_mocked_workunit():
                self.undertest.workunits.append(self.create_workunit())

            prefetch_workunit_mock.side_effect = generate_mocked_workunit

        self.undertest.prefetch_workunit = prefetch_workunit_mock

        return prefetch_workunit_mock

    def test_get_first_workunit_also_prefetches_configured_quantity(self):
        prefetch_workunit_mock = self.mock_prefetch_workunit()

        self.undertest.get_workunit()
        assert_that(prefetch_workunit_mock.call_count,
                    equal_to(self.prefetch_quantity))

    def test_get_second_workunit_prefetches_one_more(self):
        prefetch_workunit_mock = self.mock_prefetch_workunit()

        self.undertest.get_workunit()
        self.undertest.get_workunit()
        assert_that(prefetch_workunit_mock.call_count,
                    equal_to(self.prefetch_quantity + 1))

    def test_all_prefetched_doesnt_raise_no_available_work_until_all_retrieved(self):
        prefetch_workunit_mock = self.mock_prefetch_workunit(bypass_threading=True)

        workunit1 = self.create_workunit()
        workunit2 = self.create_workunit()
        workunit3 = self.create_workunit()

        self.set_workunit_provider_return_values(
            [workunit1, workunit2, workunit3, NoAvailableWorkException()])

        assert_that(self.undertest.get_workunit(), equal_to(workunit1))

        assert_that(prefetch_workunit_mock.call_count,
                    equal_to(self.prefetch_quantity))

        assert_that(self.undertest.get_workunit(), equal_to(workunit2))

        # This call to fetch will raise a NoAvailableWorkException
        assert_that(prefetch_workunit_mock.call_count,
                    equal_to(self.prefetch_quantity + 1))

        assert_that(self.undertest.get_workunit(), equal_to(workunit3))

        # Note we don't try to fetch anymore
        assert_that(prefetch_workunit_mock.call_count,
                    equal_to(self.prefetch_quantity + 1))

        self.assertRaises(NoAvailableWorkException, self.undertest.get_workunit)

    def test_prefetch_excludes_already_fetched(self):
        prefetch_workunit_mock = self.mock_prefetch_workunit(bypass_threading=True)

        workunit1 = self.create_workunit()
        workunit2 = self.create_workunit()
        workunit3 = self.create_workunit()

        self.set_workunit_provider_return_values(
            [workunit1, workunit2, workunit3, NoAvailableWorkException()])

        self.undertest.get_workunit()
        self.undertest.get_workunit()
        self.undertest.get_workunit()
        self.assertRaises(NoAvailableWorkException, self.undertest.get_workunit)

        expected_calls = [
            call(ignore_list=[]),
            call(ignore_list=[workunit1.get_filename()]),
            call(ignore_list=[workunit1.get_filename(), workunit2.get_filename()]),
            call(ignore_list=[workunit1.get_filename(), workunit2.get_filename(),
                              workunit3.get_filename()])
        ]

        self.workunit_provider.get_workunit.assert_has_calls(expected_calls)

    def test_get_workunit_no_prefetch(self):
        self.prefetch_quantity = 0
        self.workunit_provider = Mock(spec=WorkUnitProvider)
        self.undertest = PreFetchingWorkUnitProvider(self.workunit_provider,
                                                     self.prefetch_quantity)
        prefetch_workunit_mock = self.mock_prefetch_workunit(bypass_threading=True)

        workunit1 = self.create_workunit()
        workunit2 = self.create_workunit()
        workunit3 = self.create_workunit()

        self.set_workunit_provider_return_values(
            [workunit1, workunit2, workunit3, NoAvailableWorkException()])

        assert_that(self.undertest.get_workunit(), equal_to(workunit1))

        assert_that(prefetch_workunit_mock.call_count, equal_to(0))

        assert_that(self.undertest.get_workunit(), equal_to(workunit2))

        assert_that(prefetch_workunit_mock.call_count, equal_to(0))

        assert_that(self.undertest.get_workunit(), equal_to(workunit3))

        assert_that(prefetch_workunit_mock.call_count, equal_to(0))

        self.assertRaises(NoAvailableWorkException, self.undertest.get_workunit)

    def test_duplicate_workunit_ignored_(self):
        # This tests the case when 2 or more threads created back to back
        # end up retrieving the same workunit.  Only want to return it once
        # though.
        prefetch_workunit_mock = self.mock_prefetch_workunit(bypass_threading=True)

        workunit1 = self.create_workunit(1)
        workunit2 = self.create_workunit(2)
        workunit3 = self.create_workunit(2)

        self.set_workunit_provider_return_values(
            [workunit1, workunit2, workunit3, NoAvailableWorkException()])

        assert_that(self.undertest.get_workunit(), equal_to(workunit1))

        assert_that(self.undertest.get_workunit(), equal_to(workunit2))

        self.assertRaises(NoAvailableWorkException, self.undertest.get_workunit)


class WorkUnitProviderRealFilesTest(FileReadingTestCase, DirectoryCleaningTestCase):
    def setUp(self):
        working_directory = self.get_directory_to_clean()
        context = LocalDirectoryWorkingContext(working_directory)
        progress_manager = InMemoryProgressManager(context)
        parser = AstromParser()
        builder = RealsWorkUnitBuilder(parser, context, context, progress_manager)
        undertest = WorkUnitProvider(tasks.get_suffix(tasks.REALS_TASK),
                                     context,
                                     progress_manager,
                                     builder)
        self.progress_manager = progress_manager
        self.undertest = undertest

    def get_directory_to_clean(self):
        return self.get_abs_path("data/workload_testdir2")

    def get_files_to_keep(self):
        return ["candstest1.measure3.cands.astrom", "candstest2.measure3.cands.astrom",
                "candstest1.measure3.reals.astrom", "candstest2.measure3.reals.astrom",
                "realstest1.measure3.reals.astrom", "realstest2.measure3.reals.astrom"]

    def test_skip_empty_files(self):
        expected_filenames = ["realstest1.measure3.reals.astrom",
                              "realstest2.measure3.reals.astrom"]

        actual_filenames = []
        workunit1 = self.undertest.get_workunit()
        actual_filenames.append(workunit1.get_filename())
        self.progress_manager.record_done(workunit1.get_filename())

        workunit2 = self.undertest.get_workunit()
        actual_filenames.append(workunit2.get_filename())
        self.progress_manager.record_done(workunit2.get_filename())

        self.assertRaises(NoAvailableWorkException, self.undertest.get_workunit)

        assert_that(actual_filenames, contains_inanyorder(*expected_filenames))


class TracksWorkUnitBuilderTest(unittest.TestCase):
    def test_move_discovery_to_front(self):
        def mock_reading(discovery=False):
            reading = Mock(spec=SourceReading)
            reading.discovery = discovery
            return reading

        reading0 = mock_reading()
        reading1 = mock_reading(discovery=True)
        reading2 = mock_reading()
        reading3 = mock_reading()
        reading4 = mock_reading()

        data = SSOSData(Mock(),
                        [[reading0, reading1, reading2, reading3, reading4]],
                        "123456")

        builder = TracksWorkUnitBuilder(Mock(spec=SSOSParser),
                                        Mock(spec=LocalDirectoryWorkingContext),
                                        Mock(spec=LocalDirectoryWorkingContext),
                                        Mock(spec=LocalProgressManager))

        def get_readings(data):
            # Note: Track workunits only have 1 source
            return data.get_sources()[0].get_readings()

        assert_that(get_readings(data),
                    contains(reading0, reading1, reading2, reading3, reading4))

        builder.move_discovery_to_front(data)

        assert_that(get_readings(data),
                    contains(reading1, reading2, reading3, reading0, reading4))


if __name__ == '__main__':
    unittest.main()



"""
send queries to the Horizons batch query system via a web UI.  Derived from code supplied by
Wes Fraser and Michele Bannister with more smarty pants stuff added by JJ Kavelaars.
"""
import copy
import re
import requests
import scipy
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy import units
from astropy.units.quantity import Quantity
from astropy.io.ascii import Csv
from gui import logger


class Query(object):
    """
    A query object is used to build the query structure that will be sent to Horizons web system.
    """

    SERVER = 'ssd.jpl.nasa.gov'
    PROTOCOL = 'http'
    END_POINT = 'horizons_batch.cgi'

    horizons_quantities = [0, 'Astrometric RA & DEC', 'Apparent RA & DEC',
                           'Rates; RA & DEC', 'Apparent AZ & EL', 'Rates; AZ & EL',
                           'Sat. X & Y, pos. ang', 'Local app. sid. time',
                           'Airmass', 'Vis mag. & Surf Brt', 'Illuminated fraction',
                           'Defect of illumin.', 'Sat. angle separ/vis',
                           'Target angular diam.', 'Obs sub-lng & sub-lat',
                           'Sun sub-long & sub-lat', 'Sub Sun Pos. Ang & Dis',
                           'N. Pole Pos. Ang & Dis', 'Helio eclip. lon & lat',
                           'Helio range & rng rate', 'Obsrv range & rng rate',
                           'One-Way Light-Time', 'Speed wrt Sun & obsrvr',
                           'Sun-Obs-Targ ELONG ang', 'Sun-Targ-Obs PHASE ang',
                           'Targ-Obsrv-Moon/Illum%', 'Obs-Primary-Targ angl',
                           'Pos. Ang;radius & -vel', 'Orbit plane angle',
                           'Constellation ID', 'Delta-T (CT - UT)', 'Obs eclip. lon & lat',
                           'North pole RA & DEC', 'Galactic latitude', 'Local app. SOLAR time',
                           'Earth->Site lt-time', 'RA & DEC uncertainty', 'POS error ellipse',
                           'POS uncertainty (RSS)', 'Range & Rng-rate sig.', 'Doppler/delay sigmas',
                           'True anomaly angle', 'Local app. hour angle']

    default_quantities = ['Astrometric RA & DEC', 'Rates; RA & DEC',
                          'RA & DEC uncertainty', 'Vis mag. & Surf Brt',
                          'POS error ellipse']

    default_horizons_params = {'batch': "'{}'".format(1),
                               'COMMAND': "'{}'".format('Ceres'),
                               'MAKE_EPHEM': "'YES'",
                               'TABLE_TYPE': "'OBSERVER'",
                               'CAL_FORMAT': "'BOTH'",
                               'TIME_DIGITS': "'SECONDS'",
                               'ANG_FORMAT': "'DEG'",
                               'CENTER': "'568@399'",
                               'START_TIME': "'JD {}'".format(Time.now()),
                               'STOP_TIME': "'JD {}'".format(Time.now() + 10*units.day),
                               'STEP_SIZE': "'{} {}'".format(1, 'd'),
                               'QUANTITIES': None,
                               'REF_SYSTEM': "'J2000'",
                               'SKIP_DAYLT': "'NO'",
                               'EXTRA_PREC': "'NO'",
                               'R_T_S_ONLY': "'NO'",
                               'CSV_FORMAT': "'YES'"}

    def __init__(self, target, start_time, stop_time, step_size):
        """

        @type start_time: Time
        @type stop_time: Time
        @type step_size: Quantity
        @type target: str
        @return: Query
        """
        self._target = target
        self._params = copy.copy(Query.default_horizons_params)
        self._quantities = None
        self.quantities = Query.default_quantities
        self._data = None
        self._start_time = start_time
        self._stop_time = stop_time
        self._step_size = step_size
        self._center = 568
        self._nobs = None
        self._arc_length = None
        self._current_time = None

    @property
    def target(self):
        """
        The solar system body to get an ephemeris for.

        @return: str
        """
        return self._target

    @target.setter
    def target(self, target):
        if self._target != target:
            self._target = target
            self.reset()

    def reset(self):
        self._data = None

    @property
    def center(self):
        """
        The location of the observer.

        @return: str
        """
        return "'{}@399'".format(self._center)

    @center.setter
    def center(self, center):
        self._center = center

    @property
    def command(self):
        """
        The command to send to horizons, this is the same as the targeted but formatted for Horizons
        @return: str
        """
        return "'{}'".format(self.target)

    @property
    def params(self):
        """
        Build the params set for the Horizons query.  There is a default set that is overridden by matching attributes.
        @return: dict
        """
        for key in self._params:
            attr = key.lower()
            if hasattr(self, attr):
                self._params[key] = getattr(self, attr, self._params[key])
        return self._params

    @params.setter
    def params(self, **kwargs):
        """
        Set the parameters to be set on the Horizons query.

        See Query.default_horizons_params for a list of possible parameters.
        @return:
        """
        for (key, value) in enumerate(kwargs):
            self._params[key] = value

        # this is a required setting, or we have trouble with parsing.
        self._params['CSV_FORMAT'] = "'YES'"

    @property
    def quantities(self):
        """
        An array of quantities to request from Horizons for the ephemeris.  Returned as a string for Horizons query.

        @return: str
        """
        s = "{}".format(self._quantities)
        return "'{}'".format(s[1:-1])

    @quantities.setter
    def quantities(self, quantities=None):
        if quantities is None:
            quantities = []
        quantities.extend(Query.default_quantities)

        self._quantities = []
        for quantity in quantities:
            try:
                idx = int(quantity)
            except:
                idx = Query.horizons_quantities.index(quantity)
            if idx is not None and idx not in self._quantities:
                self._quantities.append(idx)

    @property
    def start_time(self):
        """
        Start time of the ephemeris, expressed as Julian Date and in string formatted expected by Horizons.

        This attribute is set to a Time object but returned as a Horizons formatted string, same for stop_time.

        @return: str
        """
        return "'JD {}'".format(self._start_time.jd)

    @start_time.setter
    def start_time(self, start_time):
        assert isinstance(start_time, Time)
        if start_time != self._start_time:
            self.reset()
            self._start_time = start_time

    @property
    def stop_time(self):
        """
        Stop time of the ephemeris, expressed as Julian Date and in string formatted expected by Horizons.

        This attribute is set to a Time object but returned as a Horizons formatted string, same for start_time.

        @return: str
        """
        return "'JD {}'".format(self._stop_time.jd)

    @stop_time.setter
    def stop_time(self, stop_time):
        assert isinstance(stop_time, Time)
        if self._stop_time != stop_time:
            self._stop_time = stop_time
            self.reset()

    @property
    def step_size(self):
        """
        Size of the step in the ephemeris.

        This is set as a Quantity object with time dimension but returned as a string in Horizons query format.

        @return: str.
        """
        s = "{:1.0f}".format(self._step_size)[:3]
        return "'{}'".format(s)

    @step_size.setter
    def step_size(self, step_size):
        assert isinstance(step_size, Quantity)
        if self._step_size != step_size:
            self._step_size = step_size
            self.reset()

    @property
    def data(self):
        """
        The a list of strings returned from the Horizons query.

        @return: list
        """
        if self._data is None:
            self.query()
        return self._data

    def query(self):
        """
        Connect to the service and make the query.

        @return:
        """
        url = '{}://{}/{}'.format(Query.PROTOCOL,
                                  Query.SERVER,
                                  Query.END_POINT)
        logger.info("Sending JPL/Hoirzons query.\n")
        response = requests.get(url, params=self.params)
        response.raise_for_status()
        self._data = []
        for line in response.iter_lines():
            self._data.append(line)


class Body(object):
    """An Horizons Ephemeris returned as a result of a query to the JPL/Horizons.

    """

    def __init__(self, name, start_time=None, stop_time=None, step_size=None, center=None):
        """

        @rtype: Ephemeris
        @param name: Body to build an ephemeris for
        @param start_time:  start time of the ephemeris
        @type start_time: Time
        @param stop_time: stop time of the ephemeris
        @type stop_time: Time
        @param step_size: size of time step for ephemeris
        @type step_size: Quantity
        """
        self.name = str(name)

        # make sure the input quantities are reasonable.
        if start_time is None:
            start_time = Time.now()
        self._start_time = Time(start_time, scale='utc')

        if stop_time is None:
            stop_time = Time.now() + 1.0 * units.day
        self._stop_time = Time(stop_time, scale='utc')
        step_size = step_size is None and 1 or step_size

        if not isinstance(step_size, Quantity):
            step_size *= units.day
        self.step_size = step_size
        if center is None:
            center = 568
        self._center = center
        self._ephemeris = None
        self._elements = None
        self._current_time = None
        self._data = None

    def __str__(self):
        s = '{:>5}: {}\n'.format('Name', self.name)
        for attr in ['a', 'e', 'Inc', 'Omega', 'omega', 'M', 'Epoch', 'n']:
            s += "{:>5}: {:<12}\n".format(attr, getattr(self, attr))
        return s

    def _reset(self):
        self._ephemeris = None
        self._elements = None
        self._nobs = None
        self._arc_length = None
        self._data = None

    @property
    def start_time(self):
        """
        Start time of the ephemeris that was retrieved from Horizons.
        @rtype:  Time
        """
        return min(self.ephemeris['Time'])

    @property
    def stop_time(self):
        """
        Stop time of the ephemeris that was retrieved from Horizons.
        @rtype: Time
        """
        return max(self.ephemeris['Time'])

    def _parse_ephemeris(self):
        """Parse the ephemeris out of the responses from Horizons and place in self._ephemeris."""

        start_of_ephemeris = '$$SOE'
        end_of_ephemeris = '$$EOE'
        start_of_failure = '!$$SOF'
        start_idx = None
        end_idx = None
        for idx, line in enumerate(self.data):
            if line.startswith(start_of_ephemeris):
                start_idx = idx
            if line.startswith(end_of_ephemeris):
                end_idx = idx

        if start_idx is None or end_idx is None:
            fail_idx = None
            for idx, line in enumerate(self.data):
                if start_of_failure in line:
                    fail_idx = idx
                    break
            msg = fail_idx is None and self.data or self.data[fail_idx-2]
            print msg
            raise ValueError(msg, "failed to build ephemeris")

        # the header of the CSV structure is 2 lines before the start_of_ephmeris
        csv_lines = [self.data[start_idx - 2]]
        csv_lines.extend(self.data[start_idx + 1: end_idx])
        csv = Csv()
        table = csv.read(csv_lines)
        try:
            table['Time'] = Time(table['Date_________JDUT'], format='jd')
        except KeyError:
            raise ValueError(self.data, "Horizons result did not contain a JD Time column, rebuild the Query.")
        self._ephemeris = table

    def _parse_elements(self):
        """Parse the elements out of the response from Horizons and place in elements object."""

        # the elements are in a '****' record at the start of file
        elements_record_started = False
        self._elements = {}
        for line in self.data:
            if line.startswith('****'):
                if elements_record_started:
                    break
                elements_record_started = True
                continue
            for part in re.findall('((\S+=)\s+(\S+))', line):
                key = part[1].strip().strip('=')
                try:
                    value = float(part[2].strip())
                except:
                    value = part[2].strip()
                self._elements[key] = value

    def _parse_obs_arc(self):
        parts = re.search('# obs: (\d+) \((\d+)-(\d+)\)', self.data)
        if parts is None:
            # Just fake some data
            self._arc_length = 30 * units.day
            self._nobs = 3
        else:
            self._arc_length = (int(parts.group(3)) - int(parts.group(2))) * units.year
            self._nobs = int(parts.group(1))

    @property
    def data(self):
        if self._data is None:
            q = Query(self.name, self._start_time, self._stop_time, self.step_size)
            q.center = self._center
            self._data = q.data
        return self._data

    @property
    def elements(self):
        """
        The orbital elements for the body, as returned by JPL/Horizons.

        @rtype: dict
        """
        if self._elements is None:
            self._parse_elements()
        return self._elements

    @property
    def ephemeris(self):
        """
        A table containing the ephemeris of the body.

        @rtype: Table
        """
        if self._ephemeris is None:
            self._parse_ephemeris()
        return self._ephemeris

    @property
    def nobs(self):
        """
        The number of observations used to determine the orbit used to build the ephmeris.

        @rtype: int
        """
        if self._nobs is None:
            self._parse_obs_arc()
        return self._nobs

    @property
    def arc_length(self):
        """
        The length of the observed arc used to determine the orbit used to build the ephemeris.

        @rtype: Quantity
        """
        if self._arc_length is None:
            self._parse_obs_arc()
        return self._arc_length

    @property
    def current_time(self):
        """
        Current time for position predictions.

        If current time is set to a value outside the bounds of the available ephemeris a new query to JPL/Horizons
        occurs.

        @rtype: Time
        @return: the time of the current ra/dec/rates selected from the ephmeris.
        """
        if self._current_time is None:
            self.current_time = self.stop_time + (self.stop_time - self.start_time)/2.0
        return self._current_time

    @current_time.setter
    def current_time(self, current_time):
        self._current_time = Time(current_time, scale='utc')
        if not (self.stop_time >= self.current_time >= self.start_time):
            self._start_time = Time(self.current_time - 10.0*units.minute)
            self._stop_time = Time(self.current_time + 10.0*units.minute)
            self.step_size = 5*units.minute
            self._reset()

    @property
    def coordinate(self):
        """
        Prediction position of the target at current_time

        @rtype: SkyCoord
        """
        ra = scipy.interp(self.current_time.jd,
                          self.ephemeris['Time'].jd,
                          self.ephemeris['R.A._(ICRF/J2000.0)']) * units.degree
        dec = scipy.interp(self.current_time.jd,
                           self.ephemeris['Time'].jd,
                           self.ephemeris['DEC_(ICRF/J2000.0)']) * units.degree
        return SkyCoord(ra, dec)

    @property
    def ra_rate(self):
        """
        Uncertainty in the prediction location.

        @rtype: Quantity angle/time
        """

        ra_rate = scipy.interp(self.current_time.jd,
                               self.ephemeris['Time'].jd,
                               self.ephemeris['dRA*cosD']) * units.arcsec / units.hour

        return ra_rate

    @property
    def dec_rate(self):
        """
        Uncertainty in the prediction location.

        @rtpye: Quantity  angle/time
        """

        dec_rate = scipy.interp(self.current_time.jd,
                                self.ephemeris['Time'].jd,
                                self.ephemeris['d(DEC)/dt']) * units.arcsec / units.hour

        return dec_rate

    @property
    def dra(self):
        """
        Uncertainty in the prediction location.

        @rtpye: Quantity  angle
        """

        dra = scipy.interp(self.current_time.jd,
                           self.ephemeris['Time'].jd,
                           self.ephemeris['RA_3sigma']) * units.arcsec

        return dra

    @property
    def ddec(self):
        """
        Uncertainty in the prediction location.

        @rtype: Quantity  angle
        """
        return scipy.interp(self.current_time.jd,
                            self.ephemeris['Time'].jd,
                            self.ephemeris['DEC_3sigma']) * units.arcsec

    @property
    def pa(self):
        """
        Plane Of Sky angle of the ra/dec uncertainty ellipse.

        @rtype: Quantity
        """
        return scipy.interp(self.current_time.jd,
                            self.ephemeris['Time'].jd,
                            self.ephemeris['Theta']) * units.degree

    def predict(self, time):
        """
        The time for which coordinate and other attributes will be computed.

        This function exists so that a horizons.Ephemeris object can be used where a orbfit.Orbfit object is used.
        @type time: Time
        @param time: Time
        """
        self.current_time = time

    @property
    def a(self):
        """
        Orbital semi-major axes
        @return: Quantity AU

        """
        return self.elements['A'] * units.AU

    @property
    def e(self):
        """
        orbital eccentricity

        @return: float
        """
        return self.elements['EC']

    @property
    def Inc(self):
        """
        orbital inclination

        @return: Quantity degree
        """
        return self.elements['IN'] * units.degree

    @property
    def M(self):
        """
        orbital Mean anomaly

        @return: Quantity degree
        """
        return self.elements['MA'] * units.degree

    @property
    def Omega(self):
        """
        orbital argument of ascending node.

        @return: Quantity degree
        """
        return self.elements['OM'] * units.degree

    @property
    def omega(self):
        """
        orbital argument of peri-focus
        @return:
        """
        return self.elements['W'] * units.degree

    @property
    def Epoch(self):
        """
        epoch of the orbital elements  in dynamical barycenter units
        @return: Time
        """
        return Time(self.elements.get('EPOCH', Time.now().jd), format='jd')

    @property
    def n(self):
        """
        mean motion
        @return: Quantity degree / day
        """
        return self.elements['N'] * units.degree / units.day

Ephmeris = Body



__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from hamcrest import assert_that, equal_to
from mock import patch

from tests.base_tests import WxWidgetTestCase
from ossos.gui.views.loading import WaitingGaugeDialog


class WaitingGaugeDialogTest(WxWidgetTestCase):
    def setUp(self):
        super(WaitingGaugeDialogTest, self).setUp()

    @patch.object(WaitingGaugeDialog, "_on_tick")
    def test_dialog(self, on_tick_mock):
        tick_time = 10
        wait_msg = "Please wait..."
        undertest = WaitingGaugeDialog(self.rootframe, wait_msg,
                                       pulse_period_ms=tick_time)

        assert_that(undertest.msg.GetLabelText(), equal_to(wait_msg))

        # TODO: test ticker? doesn't seem to fire in test case..
        # Wait for a bit more than a tick period then check ticker activated
        # time.sleep((float(tick_time) / 1000) * 2.25)
        # assert_that(on_tick_mock.call_count, greater_than(0))


if __name__ == '__main__':
    unittest.main()



__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from astropy.io import fits
from hamcrest import assert_that, equal_to
from mock import patch, ANY

from tests.base_tests import FileReadingTestCase
from ossos.downloads.cutouts.calculator import CoordinateConverter
from ossos.downloads.cutouts.source import SourceCutout
from ossos.downloads.core import ApcorData
from ossos.astrom import AstromParser


class SourceSnapshotIntegrationTest(FileReadingTestCase):
    def path(self, relative):
        return self.get_abs_path("data/image_reading/%s" % relative)

    def setUp(self):
        test_file = self.path("realstest2.measure3.reals.astrom")
        astrom_data = AstromParser().parse(test_file)

        self.reading = astrom_data.get_sources()[0].get_reading(0)

        # Load the real image
        hdulist = fits.open(self.path("cutout-1616687p.fits"))

        with open(self.path("1616687p10.apcor")) as fh:
            apcor = ApcorData.from_string(fh.read())

        # NOTE: the test image
        # vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1616687/1616687p.fits
        # has cutout: [11][1449:1199,1429:1179] and is inverted.
        #
        # Therefore the first reading at (772.13, 3334.70) in observed
        # coordinates is at
        # (NAX1 - max cutout x, NAX2 - max cutout y)
        #  = (2112 - 1449, 4644 - 1429) = (663, 3215)
        self.original_pixel_x = 663
        self.original_pixel_y = 3215

        self.original_observed_x = 772.13
        self.original_observed_y = 3334.70

        x_offset = self.original_observed_x - self.original_pixel_x
        y_offset = self.original_observed_y - self.original_pixel_y

        self.undertest = SourceCutout(self.reading, hdulist, apcor=apcor)

        assert_that(self.undertest.observed_x, equal_to(self.original_observed_x))
        assert_that(self.undertest.observed_y, equal_to(self.original_observed_y))

        assert_that(self.undertest.pixel_x, equal_to(self.original_pixel_x))
        assert_that(self.undertest.pixel_y, equal_to(self.original_pixel_y))

    @patch("ossos.wcs.xy2sky")
    def test_update_ra_dec(self, mock_xy2sky):
        new_ra = 2000.0
        new_dec = -11.0
        mock_xy2sky.return_value = (new_ra, new_dec)

        diff_x = 5.5
        diff_y = 4.5
        new_pixel_x = self.original_pixel_x + diff_x
        new_pixel_y = self.original_pixel_y + diff_y

        self.undertest.update_pixel_location((new_pixel_x, new_pixel_y))

        assert_that(self.undertest.ra, equal_to(new_ra))
        assert_that(self.undertest.dec, equal_to(new_dec))

        # Check it was called with the correct values:
        #   x, y, crpix1, crpix2, crval1, crval2, cd, pv, nord

        # These values from FITS header:

        crpix1 = 6.779623933420000E+03
        crpix2 = -3.343209710100000E+03

        crval1 = 211.828420000
        crval2 = -11.8333100000

        # CD1_1, CD1_2, CD2_1, CD2_2
        cd = [[-5.156837193510000E-05, -3.588985558218000E-07],
              [-1.674871346376000E-07, 5.178515755263000E-05]]

        # PV1_0 - PV1_10, PV2_0 - PV2_10
        pv = [[-1.909806626877E-03, 1.00815723310, 1.983393486250E-03,
               0.00000000000, 4.623172227977E-05, -1.020423075276E-05,
               2.258600565396E-05, -2.354656733463E-02, -1.671912720931E-04,
               -2.339051960031E-02, -2.903290518437E-05],
              [-7.294883106769E-04, 1.00392029973, 1.983331686156E-03,
               0.00000000000, 3.022667872225E-05, 4.509364811534E-05,
               -5.385195875663E-06, -2.333221106141E-02, -1.789051786060E-04,
               -2.348803123711E-02, -2.465552723903E-05]]

        # NORDFIT
        nord = 3

        # NOTE: the x and y passed in must be the updated OBSERVED coordinates
        mock_xy2sky.assert_called_once_with(
            new_pixel_x, new_pixel_y,
            crpix1, crpix2, crval1, crval2,
            cd, pv, nord)

    @patch("ossos.daophot.phot_mag")
    def test_get_observed_magnitude(self, mock_phot_mag):
        self.undertest.get_observed_magnitude()

        # NOTE: the x and y passed in must be the PIXEL coordinates
        mock_phot_mag.assert_called_once_with(
            ANY,
            self.original_pixel_x,
            self.original_pixel_y,
            aperture=5.0,
            sky=21.0,
            swidth=5.0,
            apcor=0.23,
            maxcount=30000.0)


if __name__ == '__main__':
    unittest.main()



from astropy import units
from baseviewer import WxMPLFitsViewer
from displayable import DisplayableImageSinglet
from interaction import Signal
from ..gui import logger
from ..downloads.cutouts.source import SourceCutout
import sys

__author__ = "David Rusk <drusk@uvic.ca>"


class SingletViewer(WxMPLFitsViewer):
    """
    Displays a single FITS image at a time.
    """

    def __init__(self, parent, display):
        logger.debug("Bilding {}".format(self))
        super(SingletViewer, self).__init__(parent, display)
        self.xy_changed = Signal()

    def _refresh_markers(self, cutout):
        self._displayables_by_cutout[cutout].clear_markers(self.ds9)
        self.mark_apertures(self.current_cutout)

    def mark_apertures(self, cutout, pixel=False):
        """

        :param cutout: A SourceCutout that is displayed
        :type cutout: SourceCutout
        :param pixel: Mark based on pixel locations or based on RA/DEC ?
        :type pixel: bool
        """
        ra = dec = None
        if not self.mark_source:
            return
        try:
            if cutout.reading.mpc_observation is not None:
                # Try using the mpc observation attached to this image.
                ra = cutout.reading.mpc_observation.coordinate.ra.to('degree').value
                dec = cutout.reading.mpc_observation.coordinate.dec.to('degree').value
                x, y, hdulist_idx = cutout.world2pix(ra, dec, usepv=True)
                (ra, dec) = cutout.pix2world(x, y, hdulist_idx, usepv=False)
            else:
                ra = cutout.reading.sky_coord.ra.to('degree').value
                dec = cutout.reading.sky_coord.dec.to('degree').value
                x, y, hdulist_idx = cutout.world2pix(ra, dec, usepv=True)
                (ra, dec) = cutout.pix2world(x, y, hdulist_idx, usepv=False)
        except Exception as ex:
            sys.stderr.write("Exception while computing location to mark aperture: {}\n".format(ex))

        try:
            # radii = (cutout.apcor.aperture, cutout.apcor.sky, cutout.apcor.swidth+cutout.apcor.sky)
            radii = [cutout.apcor.aperture, cutout.apcor.sky, cutout.apcor.swidth+cutout.apcor.sky]
        except Exception as ex:
            sys.stderr.write("Exception while accessing the aperture correction: {}\n".format(ex))
            logger.info("Exception: {0}".format(ex))
            logger.warning("Failed trying to get apcor radius values, using defaults for marking.")
            radii = [5, 25, 30]
        radii = [radius * 0.185 * units.arcsec for radius in radii]
        self._displayables_by_cutout[cutout].place_annulus(ra, dec, radii, colour='r')

    def place_marker(self, cutout, ra, dec, radius, colour, force=False):
        x, y, hdulist_index = cutout.world2pix(ra, dec, usepv=True)
        ra, dec = cutout.pix2world(x, y, hdulist_index, usepv=False)
        self._displayables_by_cutout[cutout].place_marker(ra, dec, radius, colour=colour, force=force)

    def register_xy_changed_event_handler(self, handler):
        self.xy_changed.connect(handler)

    def _create_displayable(self, cutout):
        return DisplayableImageSinglet(cutout.hdulist, self.ds9)

    def _attach_handlers(self, displayable):
        displayable.xy_changed.connect(self.xy_changed.fire)
        displayable.focus_released.connect(self.release_focus)

    def _detach_handlers(self, displayable):
        if displayable is not None:
            displayable.xy_changed.disconnect(self.xy_changed.fire)
            displayable.focus_released.disconnect(self.release_focus)



from astropy.io import ascii
from astropy.table import MaskedColumn, Table
import logging
import math
import numpy
import os
import daophot
from downloads.cutouts.downloader import ImageDownloader
import util
from downloads.cutouts.source import SourceCutout
from astrom import Observation

BRIGHT_LIMIT = 23.0
OBJECT_PLANTED = "Object.planted"
MINIMUM_BRIGHT_DETECTIONS = 5
MINIMUM_BRIGHT_FRACTION = 0.5


def measure_mags(measures):
    """
    Given a list of readings compute the magnitudes for all sources in each reading.

    @param measures: list of readings
    @return: None
    """

    image_downloader = ImageDownloader()

    observations = {}
    for measure in measures:
        for reading in measure:
            if reading.obs not in observations:
                observations[reading.obs] = {'x': [],
                                             'y': [],
                                             'source': image_downloader.download(reading, needs_apcor=True)}
            assert isinstance(reading.obs, Observation)
            observations[reading.obs]['x'].append(reading.x)
            observations[reading.obs]['y'].append(reading.y)

    for observation in observations:
        source = observations[observation]['source']
        assert isinstance(source, SourceCutout)
        source.update_pixel_location(observations[observation]['x'],
                                     observations[observation]['y'])
        hdulist_index = source.get_hdulist_idx(observation.ccdnum)
        observations[observation]['mags'] = daophot.phot(source._hdu_on_disk(hdulist_index),
                                                         observations[observation]['x'],
                                                         observations[observation]['y'],
                                                         aperture=source.apcor.aperture,
                                                         sky=source.apcor.sky,
                                                         swidth=source.apcor.swidth,
                                                         apcor=source.apcor.apcor,
                                                         zmag=source.zmag,
                                                         maxcount=30000,
                                                         extno=0)

    return observations


def match_planted(fk_candidate_observations, match_filename, bright_limit=BRIGHT_LIMIT, object_planted=OBJECT_PLANTED,
                  minimum_bright_detections=MINIMUM_BRIGHT_DETECTIONS, bright_fraction=MINIMUM_BRIGHT_FRACTION):
    """
    Using the fk_candidate_observations as input get the Object.planted file from VOSpace and match
    planted sources with found sources.

    The Object.planted list is pulled from VOSpace based on the standard file-layout and name of the
    first exposure as read from the .astrom file.

    :param fk_candidate_observations: name of the fk*reals.astrom file to check against Object.planted
    :param match_filename: a file that will contain a list of all planted sources and the matched found source
    @param minimum_bright_detections: if there are too few bright detections we raise an error.

    """

    found_pos = []
    detections = fk_candidate_observations.get_sources()
    for detection in detections:
        reading = detection.get_reading(0)
        # create a list of positions, to be used later by match_lists
        found_pos.append([reading.x, reading.y])

    # Now get the Object.planted file, either from the local FS or from VOSpace.
    objects_planted_uri = object_planted
    if not os.access(objects_planted_uri, os.F_OK):
        objects_planted_uri = fk_candidate_observations.observations[0].get_object_planted_uri()
    lines = open(objects_planted_uri).read()

    # we are changing the format of the Object.planted header to be compatible with astropy.io.ascii but
    # there are some old Object.planted files out there so we do these string/replace calls to reset those.
    new_lines = lines.replace("pix rate", "pix_rate")
    new_lines = new_lines.replace("""''/h rate""", "sky_rate")
    planted_objects_table = ascii.read(new_lines, header_start=-1, data_start=0)
    planted_objects_table.meta = None
    # The match_list method expects a list that contains a position, not an x and a y vector, so we transpose.
    planted_pos = numpy.transpose([planted_objects_table['x'].data, planted_objects_table['y'].data])
    # match_idx is an order list.  The list is in the order of the first list of positions and each entry
    # is the index of the matching position from the second list.
    (match_idx, match_fnd) = util.match_lists(numpy.array(planted_pos), numpy.array(found_pos))
    assert isinstance(match_idx, numpy.ma.MaskedArray)
    assert isinstance(match_fnd, numpy.ma.MaskedArray)
    false_positives_table = Table()
    # Once we've matched the two lists we'll need some new columns to store the information in.
    # these are masked columns so that object.planted entries that have no detected match are left 'blank'.
    new_columns = [MaskedColumn(name="measure_x", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_y", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_rate", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_angle", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_mag1", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_merr1", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_mag2", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_merr2", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_mag3", length=len(planted_objects_table), mask=True),
                   MaskedColumn(name="measure_merr3", length=len(planted_objects_table), mask=True)]
    planted_objects_table.add_columns(new_columns)
    tlength = 0
    new_columns = [MaskedColumn(name="measure_x", length=tlength, mask=True),
                   MaskedColumn(name="measure_y", length=tlength, mask=True),
                   MaskedColumn(name="measure_rate", length=0, mask=True),
                   MaskedColumn(name="measure_angle", length=0, mask=True),
                   MaskedColumn(name="measure_mag1", length=0, mask=True),
                   MaskedColumn(name="measure_merr1", length=0, mask=True),
                   MaskedColumn(name="measure_mag2", length=0, mask=True),
                   MaskedColumn(name="measure_merr2", length=0, mask=True),
                   MaskedColumn(name="measure_mag3", length=tlength, mask=True),
                   MaskedColumn(name="measure_merr3", length=tlength, mask=True)]
    false_positives_table.add_columns(new_columns)

    # We do some 'checks' on the Object.planted match to diagnose pipeline issues.  Those checks are made using just
    # those planted sources we should have detected.
    bright = planted_objects_table['mag'] < bright_limit
    n_bright_planted = numpy.count_nonzero(planted_objects_table['mag'][bright])

    measures = []
    idxs = []
    for idx in range(len(match_idx)):
        # The match_idx value is False if nothing was found.
        if not match_idx.mask[idx]:
            # Each 'source' has multiple 'readings'
            measures.append(detections[match_idx[idx]].get_readings())
            idxs.append(idx)

    observations = measure_mags(measures)

    for oidx in range(len(measures)):
        idx = idxs[oidx]
        readings = measures[oidx]
        start_jd = util.Time(readings[0].obs.header['MJD_OBS_CENTER'], format='mpc', scale='utc').jd
        end_jd = util.Time(readings[-1].obs.header['MJD_OBS_CENTER'], format='mpc', scale='utc').jd
        rate = math.sqrt((readings[-1].x - readings[0].x) ** 2 + (readings[-1].y - readings[0].y) ** 2) / (
            24 * (end_jd - start_jd))
        rate = int(rate * 100) / 100.0
        angle = math.degrees(math.atan2(readings[-1].y - readings[0].y, readings[-1].x - readings[0].x))
        angle = int(angle * 100) / 100.0
        planted_objects_table[idx]['measure_rate'] = rate
        planted_objects_table[idx]['measure_angle'] = angle
        planted_objects_table[idx]['measure_x'] = observations[readings[0].obs]['mags']["XCENTER"][oidx]
        planted_objects_table[idx]['measure_y'] = observations[readings[0].obs]['mags']["YCENTER"][oidx]
        for ridx in range(len(readings)):
            reading = readings[ridx]
            mags = observations[reading.obs]['mags']
            planted_objects_table[idx]['measure_mag{}'.format(ridx+1)] = mags["MAG"][oidx]
            planted_objects_table[idx]['measure_merr{}'.format(ridx+1)] = mags["MERR"][oidx]

    # for idx in range(len(match_fnd)):
    #     if match_fnd.mask[idx]:
    #         measures = detections[idx].get_readings()
    #         false_positives_table.add_row()
    #         false_positives_table[-1] = measure_mags(measures, false_positives_table[-1])

    # Count an object as detected if it has a measured magnitude in the first frame of the triplet.
    n_bright_found = numpy.count_nonzero(planted_objects_table['measure_mag1'][bright])
    # Also compute the offset and standard deviation of the measured magnitude from that planted ones.
    offset = numpy.mean(planted_objects_table['mag'][bright] - planted_objects_table['measure_mag1'][bright])
    try:
        offset = "{:5.2f}".format(offset)
    except:
        offset = "indef"

    std = numpy.std(planted_objects_table['mag'][bright] - planted_objects_table['measure_mag1'][bright])
    try:
        std = "{:5.2f}".format(std)
    except:
        std = "indef"

    if os.access(match_filename, os.R_OK):
        fout = open(match_filename, 'a')
    else:
        fout = open(match_filename, 'w')

    fout.write("#K {:10s} {:10s}\n".format("EXPNUM", "FWHM"))
    for measure in detections[0].get_readings():
        fout.write('#V {:10s} {:10s}\n'.format(measure.obs.header['EXPNUM'], measure.obs.header['FWHM']))

    fout.write("#K ")
    for keyword in ["RMIN", "RMAX", "ANGLE", "AWIDTH"]:
        fout.write("{:10s} ".format(keyword))
    fout.write("\n")

    fout.write("#V ")
    for keyword in ["RMIN", "RMAX", "ANGLE", "AWIDTH"]:
        fout.write("{:10s} ".format(fk_candidate_observations.sys_header[keyword]))
    fout.write("\n")

    fout.write("#K ")
    for keyword in ["NBRIGHT", "NFOUND", "OFFSET", "STDEV"]:
        fout.write("{:10s} ".format(keyword))
    fout.write("\n")
    fout.write("#V {:<10} {:<10} {:<10} {:<10}\n".format(n_bright_planted,
                                                         n_bright_found,
                                                         offset,
                                                         std))

    try:
        writer = ascii.FixedWidth
        # add a hash to the start of line that will have header columns: for JMP
        fout.write("#")
        fout.flush()
        ascii.write(planted_objects_table, output=fout, Writer=writer, delimiter=None)
        if len(false_positives_table) > 0:
            with open(match_filename+".fp", 'a') as fpout:
                fpout.write("#")
                ascii.write(false_positives_table, output=fpout, Writer=writer, delimiter=None)
    except Exception as e:
        logging.error(str(e))
        raise e
    finally:
        fout.close()

    # Some simple checks to report a failure how we're doing.
    if n_bright_planted < minimum_bright_detections:
        raise RuntimeError(1, "Too few bright objects planted.")

    if n_bright_found / float(n_bright_planted) < bright_fraction:
        raise RuntimeError(2, "Too few bright objects found.")

    return "{} {} {} {}".format(n_bright_planted, n_bright_found, offset, std)




import os
import random
import re
import threading
from glob import glob

from .collections import StatefulCollection
from .exceptions import (NoAvailableWorkException, SourceNotNamedException)
from .. import events
from .. import logger
from .. import tasks
from ..progress import FileLockedException
from ...astrom import StreamingAstromWriter, Source, SourceReading, StreamingVettingWriter
from ...orbfit import Orbfit

__author__ = "David Rusk <drusk@uvic.ca>"


class WorkUnit(object):
    """
    A unit of work to be processed, associated with the data from a single
    input file.
    """

    def __init__(self, filename,
                 parsed_data,
                 progress_manager,
                 output_context,
                 dry_run=False):
        self.filename = filename
        self.data = parsed_data
        self.progress_manager = progress_manager
        self.output_context = output_context
        self.dry_run = dry_run

        self.sources = StatefulCollection(parsed_data.get_sources())

        self.readings_by_source = {}
        for source in self.sources:
            self.readings_by_source[source] = StatefulCollection(source.get_readings())

        self.processed_items = set()
        self._mark_previously_processed_items()

        if self.is_current_item_processed():
            self.next_item()

        self.finished_callbacks = []

        self._unlocked = False

    def register_finished_callback(self, callback):
        self.finished_callbacks.append(callback)

    def next_source(self):
        self.get_sources().next()

    def previous_source(self):
        self.get_sources().previous()

    def next_obs(self):
        self.get_current_source_readings().next()

    def previous_obs(self):
        self.get_current_source_readings().previous()

    def next_item(self):
        raise NotImplementedError()

    def accept_current_item(self):
        self.process_current_item()

    def reject_current_item(self):
        self.process_current_item()

    def process_current_item(self):
        self.processed_items.add(self.get_current_item())

        if self.is_current_source_finished():
            self.progress_manager.record_index(self.get_filename(),
                                               self.get_current_source_number())
            self.get_writer().flush()

        if self.is_finished():
            self.progress_manager.record_done(self.get_filename())
            self.unlock()

            for callback in self.finished_callbacks:
                callback(self.get_results_file_paths())

            self._close_writers()

    def get_current_item(self):
        raise NotImplementedError()

    def get_current_item_index(self):
        raise NotImplementedError()

    def is_item_processed(self, item):
        return item in self.processed_items

    def is_current_item_processed(self):
        return self.is_item_processed(self.get_current_item())

    def is_source_finished(self, source):
        raise NotImplementedError()

    def is_current_source_finished(self):
        return self.is_source_finished(self.get_current_source())

    def get_filename(self):
        return self.filename

    def get_data(self):
        return self.data

    def get_sources(self):
        return self.sources

    def get_unprocessed_sources(self):
        unprocessed_sources = []
        for source in self.get_sources():
            if not self.is_source_finished(source):
                unprocessed_sources.append(source)

        return unprocessed_sources

    def get_source_count(self):
        return len(self.get_sources())

    def get_current_source(self):
        source = self.get_sources().get_current_item()
        assert isinstance(source, Source)
        return source

    def get_current_source_number(self):
        return self.get_sources().get_index()

    def get_current_source_readings(self):
        return self.readings_by_source[self.get_current_source()]

    def get_obs_count(self):
        return len(self.get_current_source_readings())

    def get_current_reading(self):
        """
        :return: SourceReading
        :rtype: SourceReading
        """
        return self.get_current_source_readings().get_current_item()

    def get_current_obs_number(self):
        return self.get_current_source_readings().get_index()

    def get_writer(self):
        raise NotImplementedError()

    def get_results_file_paths(self):
        raise NotImplementedError()

    def is_finished(self):
        return len(self._get_item_set() - self.processed_items) == 0

    def is_apcor_needed(self):
        raise NotImplementedError()

    def unlock(self):
        if not self._unlocked:
            self.progress_manager.unlock(self.get_filename(), async=True)
            self._unlocked = True

    def _get_item_set(self):
        raise NotImplementedError()

    def _mark_previously_processed_items(self):
        pass

    def _close_writers(self):
        pass


class RealsWorkUnit(WorkUnit):
    """
    A unit of work when performing the process reals task.
    """

    def __init__(self,
                 filename,
                 parsed_data,
                 progress_manager,
                 output_context,
                 dry_run=False):
        super(RealsWorkUnit, self).__init__(
            filename,
            parsed_data,
            progress_manager,
            output_context,
            dry_run=dry_run)
        self._writers = {}

    def next_item(self):
        assert not self.is_finished()

        self.next_obs()
        while self.is_current_item_processed():
            self._next_sequential_item()

    def _next_sequential_item(self):
        """
        Go to the next item in the 'ideal' processing sequence.
        """
        if self.get_current_source_readings().is_on_last_item():
            self.next_source()
        else:
            self.next_obs()

    def get_current_item(self):
        return self.get_current_reading()

    def get_current_item_index(self):
        return (self.get_sources().get_index() * self.get_obs_count() +
                self.get_current_source_readings().get_index())

    def is_source_finished(self, source):
        for reading in source.get_readings():
            if reading not in self.processed_items:
                return False

        return True

    def is_apcor_needed(self):
        return True

    def get_writer(self):
        filename = self.get_output_filename(self.get_current_source())
        if filename in self._writers:
            return self._writers[filename]

        writer = self._create_writer(filename)
        self._writers[filename] = writer
        return writer

    def get_results_file_paths(self):
        return [self.output_context.get_full_path(filename)
                for filename in self._writers]

    def get_output_filename(self, source):
        if not source.has_provisional_name():
            raise SourceNotNamedException(source)

        return ".".join((os.path.basename(self.filename),
                         source.get_provisional_name(),
                         "mpc"))

    def _create_writer(self, filename):
        # NOTE: this import is only here so that we don't load up secondary
        # dependencies (like astropy) used in MPCWriter when they are not
        # needed (i.e. cands task).  This is to help reduce the application
        # startup time.
        from ...mpc import MPCWriter

        return MPCWriter(self.output_context.open(filename),
                         auto_flush=False)

    def _get_item_set(self):
        all_readings = set()
        for readings in self.readings_by_source.itervalues():
            all_readings.update(readings)
        return all_readings

    def _mark_previously_processed_items(self):
        processed_indices = self.progress_manager.get_processed_indices(self.get_filename())
        for index in processed_indices:
            for reading in self.get_sources()[index].get_readings():
                self.processed_items.add(reading)

    def _close_writers(self):
        for writer in self._writers.values():
            writer.close()


class CandidatesWorkUnit(WorkUnit):
    """
    A unit of work when performing the process candidates task.
    """

    def __init__(self,
                 filename,
                 parsed_data,
                 progress_manager,
                 output_context,
                 dry_run=False):
        super(CandidatesWorkUnit, self).__init__(
            filename,
            parsed_data,
            progress_manager,
            output_context,
            dry_run=dry_run)

        self._writer = None

    def next_item(self):
        assert not self.is_finished()

        self.next_source()
        while self.is_current_item_processed():
            self.next_source()

    def get_current_item(self):
        return self.get_current_source()

    def get_current_item_index(self):
        return self.get_sources().get_index()

    def is_source_finished(self, source):
        return source in self.processed_items

    def is_apcor_needed(self):
        return False

    def get_writer(self):
        if self._writer is None:
            self._writer = self._create_writer()

        return self._writer

    def get_results_file_paths(self):
        if self._writer is None:
            return []

        return [self.output_context.get_full_path(self._writer.get_filename())]

    def get_output_filename(self):
        return self.get_filename().replace(tasks.get_suffix(tasks.CANDS_TASK),
                                           tasks.get_suffix(tasks.REALS_TASK))

    def _create_writer(self):
        filename = self.get_output_filename()
        return StreamingAstromWriter(self.output_context.open(filename),
                                     self.data.sys_header)

    def _get_item_set(self):
        return set(self.sources)

    def _mark_previously_processed_items(self):
        processed_indices = self.progress_manager.get_processed_indices(self.get_filename())
        for index in processed_indices:
            self.processed_items.add(self.get_sources()[index])

    def _close_writers(self):
        if self._writer is not None:
            self._writer.close()


class VettingWorkUnit(CandidatesWorkUnit):
    """
    A unit of work when performing the process candidates task.
    """

    def _create_writer(self):
        filename = self.get_output_filename()
        return StreamingVettingWriter(self.output_context.open(filename))

    def get_output_filename(self):
        return self.get_filename().replace(tasks.get_suffix(tasks.VETTING_TASK),
                                           tasks.get_suffix(tasks.EXAMINE_TASK))


class TracksWorkUnit(WorkUnit):
    """
    A unit of work when performing the process track task.
    """

    def __init__(self,
                 builder,
                 filename,
                 parsed_data,
                 progress_manager,
                 output_context,
                 dry_run=False):
        super(TracksWorkUnit, self).__init__(
            filename,
            parsed_data,
            progress_manager,
            output_context,
            dry_run=dry_run)

        self.builder = builder
        self._writer = None
        self._ssos_queried = False
        self._comparitors = {}

    def print_orbfit_info(self):
        # TODO: this should not be here.
        orb = Orbfit(self.get_writer().get_chronological_buffered_observations())
        print orb.residuals()
        print orb

    def query_ssos(self):
        """
        Use the MPC file that has been built up in processing this work
        unit to generate another workunit.
        """
        self._ssos_queried = True
        mpc_filename = self.save()
        return self.builder.build_workunit(mpc_filename)

    def save(self):
        """
        Update the SouceReading information for the currently recorded observations and then flush those to a file.
        @return: mpc_filename of the resulting save.
        """
        self.get_writer().flush()
        mpc_filename = self.output_context.get_full_path(self.get_writer().get_filename())
        self.get_writer().close()
        self._writer = None
        return mpc_filename

    def is_finished(self):
        return self._ssos_queried or self.get_source_count() == 0 or super(TracksWorkUnit, self).is_finished()

    def next_item(self):
        assert not self.is_finished()

        self.next_obs()
        while self.is_current_item_processed():
            self._next_sequential_item()

    def _next_sequential_item(self):
        """
        Go to the next item to process.
        """
        if self.get_current_source_readings().is_on_last_item():
            self.next_source()
        else:
            self.next_obs()

    def get_current_item(self):
        return self.get_current_reading()

    def get_current_item_index(self):
        return (self.get_sources().get_index() * self.get_obs_count() +
                self.get_current_source_readings().get_index())

    def is_source_finished(self, source):
        for reading in source.get_readings():
            if reading not in self.processed_items:
                return False

        return True

    def is_apcor_needed(self):
        return True

    def get_writer(self):
        """
        Get a writer.

        This method also makes the output filename be the same as the .track file but with .mpc.
        (Currently only works on local filesystem)
        :rtype MPCWriter
        """
        if self._writer is None:
            suffix = tasks.get_suffix(tasks.TRACK_TASK)
            try:
                base_name = re.search("(?P<base_name>.*?)\.\d*{}".format(suffix), self.filename).group('base_name')
            except:
                base_name = os.path.splitext(self.filename)[0]
            mpc_filename_pattern = self.output_context.get_full_path(
                "{}.?{}".format(base_name, suffix))
            mpc_file_count = len(glob(mpc_filename_pattern))
            mpc_filename = self.output_context.get_full_path(
                "{}.{}{}".format(base_name, mpc_file_count, suffix))
            self._writer = self._create_writer(mpc_filename)

        return self._writer

    def get_results_file_paths(self):
        if self._writer is None:
            return []

        return [self.output_context.get_full_path(self._writer.get_filename())]

    def _create_writer(self, filename):
        # NOTE: this import is only here so that we don't load up secondary
        # dependencies (like astropy) used in MPCWriter when they are not
        # needed (i.e. cands task).  This is to help reduce the application
        # startup time.
        from ...mpc import MPCWriter

        writer = MPCWriter(self.output_context.open(filename),
                           auto_flush=False, auto_discovery=False)

        for rawname in self.data.mpc_observations:
            writer.write(self.data.mpc_observations[rawname])
        
        return writer

    def _get_item_set(self):
        all_readings = set()
        for readings in self.readings_by_source.itervalues():
            all_readings.update(readings)
        return all_readings

    def _mark_previously_processed_items(self):
        processed_indices = self.progress_manager.get_processed_indices(self.get_filename())
        for index in processed_indices:
            for reading in self.get_sources()[index].get_readings():
                self.processed_items.add(reading)

    def _close_writers(self):
        if self._writer is not None:
            self._writer.close()


class TargetWorkUnit(TracksWorkUnit):

    def __init__(self,
                 builder,
                 filename,
                 parsed_data,
                 progress_manager,
                 output_context,
                 dry_run=False):
        super(TracksWorkUnit, self).__init__(
            filename,
            parsed_data,
            progress_manager,
            output_context,
            dry_run=dry_run)


class WorkUnitProvider(object):
    """
    Obtains new units of work for the application.
    """

    def __init__(self,
                 taskid,
                 directory_context,
                 progress_manager,
                 builder,
                 randomize=False,
                 name_filter=None):
        self.taskid = taskid
        self.directory_context = directory_context
        self.progress_manager = progress_manager
        self.builder = builder
        self.randomize = randomize
        self.name_filter = name_filter
        self._done = []
        self._already_fetched = []

    @property
    def directory(self):
        """
        The directory that workunits are being acquired from.
        """
        return self.directory_context.directory

    def _filter(self, filename):
        """
        return 'true' if filename doesn't match name_filter regex and should be filtered out of the list.
        @param filename:
        @return:
        """
        return self.name_filter is not None and re.search(self.name_filter, filename) is None

    def get_workunit(self, ignore_list=None):
        """
        Gets a new unit of work.

        Args:
          ignore_list: list(str)
            A list of filenames which should be ignored.  Defaults to None.

        Returns:
          new_workunit: WorkUnit
            A new unit of work that has not yet been processed.  A lock on
            it has been acquired.

        Raises:
          NoAvailableWorkException
            There is no more work available.
        """
        if ignore_list is None:
            ignore_list = []

        potential_files = self.get_potential_files(ignore_list)

        while len(potential_files) > 0:
            potential_file = self.select_potential_file(potential_files)
            potential_files.remove(potential_file)
            if self._filter(potential_file):
                continue

            if self.directory_context.get_file_size(potential_file) == 0:
                continue

            if self.progress_manager.is_done(potential_file):
                self._done.append(potential_file)
                continue
            else:
                try:
                    self.progress_manager.lock(potential_file)
                except FileLockedException:
                    continue

                self._already_fetched.append(potential_file)
                return self.builder.build_workunit(
                    self.directory_context.get_full_path(potential_file))

        logger.info("No eligible workunits remain to be fetched.")

        raise NoAvailableWorkException()

    def get_potential_files(self, ignore_list):
        """
        Get a listing of files for the appropriate task which may or may
        not be locked and/or done.
        """
        exclude_prefix = self.taskid == tasks.suffixes.get(tasks.REALS_TASK, '') and 'fk' or None
        filenames = [filename for filename in
                     self.directory_context.get_listing(self.taskid, exclude_prefix=exclude_prefix)
                     if filename not in ignore_list and
                     filename not in self._done and
                     filename not in self._already_fetched]

        # if the extension is .mpc. then we look for the largest numbered MPC file.
        # look for the largest numbered MPC file only.
        if self.taskid == tasks.suffixes.get(tasks.TRACK_TASK, ''):
            basenames = {}
            for filename in filenames:
                fullname = os.path.splitext(filename)[0]
                if fullname in basenames:
                    continue
                basename = os.path.splitext(fullname)[0]
                # only do the 'maximum' search when the 2nd extension is an integer value
                try:
                    idx = int(filename.split('.')[-2])
                    if idx > basenames.get(basename, 0):
                        basenames[basename] = idx
                except:
                    # since we failed, just keep the file in the list
                    basenames[fullname] = ''
            filenames = []
            for basename in basenames:
                # sometimes the version is empty, so no '.' is needed
                version = basenames[basename]
                version = len(str(version)) > 0 and ".{}".format(version) or version
                filenames.append("{}{}{}".format(basename, version, self.taskid))
                # print basename, basenames[basename], filenames[-1]
        return filenames

    def select_potential_file(self, potential_files):
        if self.randomize:
            # Don't want predictable patterns in the order we get work units.
            return random.choice(potential_files)
        else:
            return potential_files[0]

    def shutdown(self):
        pass


class PreFetchingWorkUnitProvider(object):
    def __init__(self, workunit_provider, prefetch_quantity, image_manager):
        self.workunit_provider = workunit_provider
        self.prefetch_quantity = prefetch_quantity
        self.image_manager = image_manager

        self.fetched_files = []
        self.workunits = []

        self._threads = []
        self._all_fetched = False

    @property
    def directory(self):
        """
        The directory that workunits are being acquired from.
        """
        return self.workunit_provider.directory

    def get_workunit(self):
        if self._all_fetched and len(self.workunits) == 0:
            raise NoAvailableWorkException()

        if len(self.workunits) > 0:
            workunit = self.workunits.pop(0)
        else:
            workunit = self.workunit_provider.get_workunit(
                ignore_list=self.fetched_files)
            self.fetched_files.append(workunit.get_filename())

        self.trigger_prefetching()
        logger.debug("Returning {}".format(workunit))
        return workunit

    def trigger_prefetching(self):
        if self._all_fetched:
            return

        num_to_fetch = self.prefetch_quantity - len(self.workunits)

        if num_to_fetch < 0:
            # Prefetch quantity can be 0
            num_to_fetch = 0

        while num_to_fetch > 0:
            if self._all_fetched:
                return

            self.prefetch_workunit()
            num_to_fetch -= 1

    def prefetch_workunit(self):
        thread = threading.Thread(target=self._do_prefetch_workunit)
        self._threads.append(thread)
        thread.start()

    def _do_prefetch_workunit(self):
        try:
            workunit = self.workunit_provider.get_workunit(
                ignore_list=self.fetched_files)
            filename = workunit.get_filename()

            # 2 or more threads created back to back could end up
            # retrieving the same workunit.  Only keep one of them.
            if filename not in self.fetched_files:
                self.fetched_files.append(filename)
                self.workunits.append(workunit)
                self.image_manager.download_singlets_for_workunit(workunit)

                logger.info("%s was prefetched." % filename)

        except NoAvailableWorkException:
            self._all_fetched = True

    def shutdown(self):
        # Make sure all threads are finished so that no more locks are
        # acquired
        for thread in self._threads:
            thread.join()

        for workunit in self.workunits:
            workunit.unlock()


class WorkUnitBuilder(object):
    """
    Used to construct a WorkUnit with its necessary components.
    """

    def __init__(self, parser, input_context, output_context, progress_manager,
                 dry_run=False):
        logger.debug("Setting up Work Unit Builder.")
        self.parser = parser
        self.input_context = input_context
        self.output_context = output_context
        self.progress_manager = progress_manager
        self.dry_run = dry_run

    def build_workunit(self, input_fullpath):
        try:
            parsed_data = self.parser.parse(input_fullpath)
            logger.debug("Parsed %s (%d sources)" %
                         (input_fullpath, parsed_data.get_source_count()))

            input_filename = os.path.split(input_fullpath)[1]
            return self._do_build_workunit(
                input_filename,
                parsed_data,
                self.progress_manager,
                self.output_context,
                self.dry_run)

        except AssertionError as e:
            logger.critical(str(e))
            events.send(events.NO_AVAILABLE_WORK)

    def _do_build_workunit(self,
                           filename,
                           data,
                           progress_manager,
                           output_context,
                           dry_run):
        raise NotImplementedError()


class TracksWorkUnitBuilder(WorkUnitBuilder):
    """
    Used to construct a WorkUnit for doing 'Track'
    """

    def __init__(self, parser, input_context, output_context, progress_manager,
                 dry_run=False):
        super(TracksWorkUnitBuilder, self).__init__(
            parser, input_context, output_context, progress_manager,
            dry_run=dry_run
        )

    @staticmethod
    def get_readings(data):
        # Note: Track workunits only have 1 source
        return data.get_sources()[0].get_readings()

    @staticmethod
    def set_readings(data, readings):
        # Note: Track workunits only have 1 source
        data.get_sources()[0].readings = readings

    def get_discovery_index(self, data):
        # Note: there should only be one reading marked "discovery" amongst
        # the data.
        for i, reading in enumerate(self.get_readings(data)):
            if reading.discovery:
                return i
        logger.warning("No discovery index found in track workunit.")
        return None

    def move_discovery_to_front(self, data):
        """
        Moves the discovery triplet to the front of the reading list.
        Leaves everything else in the same order.
        """
        readings = self.get_readings(data)
        discovery_index = self.get_discovery_index(data)

        reordered_readings = (readings[discovery_index:discovery_index + 3] +
                              readings[:discovery_index] +
                              readings[discovery_index + 3:])

        self.set_readings(data, reordered_readings)

    def _do_build_workunit(self,
                           filename,
                           data,
                           progress_manager,
                           output_context,
                           dry_run):

        # Doesn't work... TODO: FixMe
        # self.move_discovery_to_front(data)

        return TracksWorkUnit(self, filename, data, progress_manager, output_context, dry_run=dry_run)


class TargetWorkUnitBuilder(TracksWorkUnitBuilder):

    def __init__(self, parser, input_context, output_context, progress_manager,
                 dry_run=False):
        super(TracksWorkUnitBuilder, self).__init__(
            parser, input_context, output_context, progress_manager,
            dry_run=dry_run)

    @staticmethod
    def get_readings(data):
        # Note: Track workunits only have 1 source
        return data.get_sources()[0].get_readings()

    @staticmethod
    def set_readings(data, readings):
        # Note: Track workunits only have 1 source
        data.get_sources()[0].readings = readings

    def get_discovery_index(self, data):
        return None

    def move_discovery_to_front(self, data):
        pass

    def _do_build_workunit(self,
                           filename,
                           data,
                           progress_manager,
                           output_context,
                           dry_run):

        # Doesn't work... TODO: FixMe
        # self.move_discovery_to_front(data)

        return TracksWorkUnit(self, filename, data, progress_manager, output_context, dry_run=dry_run)


class RealsWorkUnitBuilder(WorkUnitBuilder):
    """
    Used to construct a WorkUnit with its necessary components.
    Constructs RealsWorkUnits for the process reals task.
    """

    def __init__(self, parser, input_context, output_context, progress_manager,
                 dry_run=False):
        super(RealsWorkUnitBuilder, self).__init__(
            parser, input_context, output_context, progress_manager,
            dry_run=dry_run)

    def _do_build_workunit(self,
                           filename,
                           data,
                           progress_manager,
                           output_context,
                           dry_run):
        return RealsWorkUnit(
            filename, data, progress_manager, output_context,
            dry_run=dry_run)


class CandidatesWorkUnitBuilder(WorkUnitBuilder):
    """
    Used to construct a WorkUnit with its necessary components.
    Constructs CandidatesWorkUnits for the process candidates task.
    """

    def __init__(self, parser, input_context, output_context, progress_manager,
                 dry_run=False):
        super(CandidatesWorkUnitBuilder, self).__init__(
            parser, input_context, output_context, progress_manager,
            dry_run=dry_run)

    def _do_build_workunit(self,
                           filename,
                           data,
                           progress_manager,
                           output_context,
                           dry_run):
        return CandidatesWorkUnit(
            filename, data, progress_manager, output_context,
            dry_run=dry_run)


class VettingWorkUnitBuilder(WorkUnitBuilder):
    """
    Used to construct a WorkUnit with its necessary components.
    Constructs CandidatesWorkUnits for the process candidates task.
    """

    def __init__(self, parser, input_context, output_context, progress_manager,
                 dry_run=False):
        super(VettingWorkUnitBuilder, self).__init__(
            parser, input_context, output_context, progress_manager,
            dry_run=dry_run)

    def _do_build_workunit(self,
                           filename,
                           data,
                           progress_manager,
                           output_context,
                           dry_run):
        return VettingWorkUnit(
            filename, data, progress_manager, output_context,
            dry_run=dry_run)


class ExamineWorkUnitBuilder(WorkUnitBuilder):
    """
    Used to construct a WorkUnit with its necessary components.
    Constructs CandidatesWorkUnits for the process candidates task.
    """

    def __init__(self, parser, input_context, output_context, progress_manager,
                 dry_run=False):
        super(ExamineWorkUnitBuilder, self).__init__(
            parser, input_context, output_context, progress_manager,
            dry_run=dry_run)

    def _do_build_workunit(self,
                           filename,
                           data,
                           progress_manager,
                           output_context,
                           dry_run):
        return ExamineWorkUnit(
            filename, data, progress_manager, output_context,
            dry_run=dry_run)


class ExamineWorkUnit(RealsWorkUnit):
    pass


from matplotlib import cm
from matplotlib import rcParams
import math 

def setFigForm():
    """set the rcparams to EmulateApJ columnwidth=245.26 pts """

    fig_width_pt = 245.26*2
    inches_per_pt = 1.0/72.27
    golden_mean = (math.sqrt(5.)-1.0)/2.0
    fig_width = fig_width_pt*inches_per_pt
    fig_height = fig_width*golden_mean
    fig_size = [1.5*fig_width, fig_height]
    
    params = {'backend': 'ps',
              'axes.labelsize': 12,
              'text.fontsize': 12,
              'legend.fontsize': 7,
              'xtick.labelsize': 11,
              'ytick.labelsize': 11,
              'text.usetex': True,
              'font.family': 'serif',
              'font.serif': 'Times',
              'image.aspect': 'auto',
              'figure.subplot.left': 0.1,
              'figure.subplot.bottom': 0.1,
              'figure.subplot.hspace': 0.25,
              'figure.figsize': fig_size}
    
    rcParams.update(params)




from ossos.gui import config

__author__ = "David Rusk <drusk@uvic.ca>"

try:
   import pyds9 as ds9
except:
   import ds9
   ds9.DS9 = ds9.ds9

from ...fitsviewer.singletviewer import SingletViewer
from ...fitsviewer.tripletviewer import TripletViewer
from ...gui import logger


class ImageViewManager(object):
    def __init__(self, mainframe, zoom=1):
        # Note: the figure we pass in is just a temporary placeholder.
        # 'Displayable Items' provide their own figure which the canvas can
        # be made to use, but it also requires one on its creation.
        logger.debug("Building {}".format(self))
        self._ds9 = None
        self.zoom = zoom
        self._singlet_viewer = SingletViewer(mainframe.main_panel, display=self.ds9)
        self._triplet_viewer = TripletViewer(mainframe.main_panel, display=self.ds9)

        self._image_viewer = self._singlet_viewer

    def set_ds9(self, level="PREF"):
        """
        Set the default values on the ds9 display.
        """
        self.set_zoom()
        ds9_settings = config.read("DS9."+level)
        for key in ds9_settings.keys():
            value = ds9_settings[key]
            cmd = key.replace("_", " ")
            self.ds9.set("{} {}".format(cmd, value))

    def set_zoom(self):
        cmd = "zoom {}".format(self.zoom)
        self.ds9.set(cmd)

    def _new_frame(self):
        self.ds9.set('frame new')

    @property
    def ds9(self):
        if not self._ds9:
            # start xpans if needed
            logger.debug("Starting XPANS")
            ds9.ds9_xpans()

            logger.debug("Starting DS9")
            # start ds9 if need, or connect to existing
            try:
                self._ds9 = ds9.DS9(target='validate', start=False)
            except ValueError as ve:
                cnt = 0
                while cnt < 10:
                    cnt += 1
                    try:
                        self._ds9 = ds9.DS9(target='validate', start=True, wait=10, verify=True)
                        self.set_ds9(level="INIT")
                        self.set_ds9(level="PREF")
                        break
                    except ValueError as ve:
                        pass
            finally:
                self._ds9.set("frame delete all")
        self._ds9.reset_preferences = self.set_ds9
        self._ds9.new_frame = self._new_frame
        return self._ds9

    @property
    def image_viewer(self):
        """

        @rtype: SingletViewer
        """
        return self._image_viewer

    @image_viewer.setter
    def image_viewer(self, viewer):
        self._image_viewer = viewer

    def use_singlets(self):
        if self.image_viewer == self._triplet_viewer:
            self.image_viewer = self._singlet_viewer

    def use_triplets(self):
        if self.image_viewer == self._singlet_viewer:
            self.image_viewer = self._triplet_viewer



__author__ = "David Rusk <drusk@uvic.ca>"

import getpass
import unittest

from hamcrest import assert_that, equal_to, contains, contains_inanyorder, has_length
from mock import patch

from tests.base_tests import FileReadingTestCase
from ossos import storage
from ossos.gui.context import VOSpaceWorkingContext
from ossos.gui import progress
from ossos.gui.progress import VOSpaceProgressManager, FileLockedException, RequiresLockException

BASE_TEST_DIR = "vos:OSSOS/integration_tests/"
PROTOTYPE_FILE = "data/prototype"
PERSISTENCE_TEST_DIR = BASE_TEST_DIR + "persistence_tests"

TEST_FILE_1 = "1.cands.astrom"
TEST_FILE_2 = "2.reals.astrom"
TEST_FILE_3 = "3.cands.astrom"

TEST_USER = "testuser"


class AbstractVOSpaceProgressManagerTestCase(object):
    # Do not run this class directly; run its subclasses

    def create_vofile(self, destination):
        # Just copy a prototype file until I figure out how to do this
        # properly
        storage.copy(self.get_abs_path(PROTOTYPE_FILE), destination)

    def setUp(self):
        self.main_user_id = "main_user"
        self.context = VOSpaceWorkingContext(PERSISTENCE_TEST_DIR)
        self.undertest = VOSpaceProgressManager(
            self.context, userid=self.main_user_id,
            track_partial_progress=self._tracks_partial_progress())

        self.create_vofile(self.context.get_full_path(TEST_FILE_1))
        self.create_vofile(self.context.get_full_path(TEST_FILE_2))
        self.create_vofile(self.context.get_full_path(TEST_FILE_3))

    def get_uri(self, filename):
        return self.context.get_full_path(filename)

    def get_property(self, filename, property):
        return storage.get_property(self.get_uri(filename), property)

    def has_property(self, filename, property):
        return storage.has_property(self.get_uri(filename), property)

    def tearDown(self):
        for filename in self.context.listdir():
            storage.delete_uri(self.context.get_full_path(filename))

    def create_independent_manager(self, userid="test_user"):
        return VOSpaceProgressManager(
            self.context, userid=userid,
            track_partial_progress=self._tracks_partial_progress())

    def _tracks_partial_progress(self):
        raise NotImplementedError()

    def test_is_done_record_done(self):
        assert_that(self.undertest.is_done(TEST_FILE_1), equal_to(False))
        assert_that(self.undertest.is_done(TEST_FILE_2), equal_to(False))
        assert_that(self.undertest.is_done(TEST_FILE_3), equal_to(False))

        self.undertest.lock(TEST_FILE_2)
        self.undertest.record_done(TEST_FILE_2)
        self.undertest.unlock(TEST_FILE_2)
        assert_that(self.undertest.is_done(TEST_FILE_1), equal_to(False))
        assert_that(self.undertest.is_done(TEST_FILE_2), equal_to(True))
        assert_that(self.undertest.is_done(TEST_FILE_3), equal_to(False))

        self.undertest.lock(TEST_FILE_3)
        self.undertest.record_done(TEST_FILE_3)
        self.undertest.unlock(TEST_FILE_3)
        assert_that(self.undertest.is_done(TEST_FILE_1), equal_to(False))
        assert_that(self.undertest.is_done(TEST_FILE_2), equal_to(True))
        assert_that(self.undertest.is_done(TEST_FILE_3), equal_to(True))

        self.undertest.lock(TEST_FILE_1)
        self.undertest.record_done(TEST_FILE_1)
        self.undertest.unlock(TEST_FILE_1)
        assert_that(self.undertest.is_done(TEST_FILE_1), equal_to(True))
        assert_that(self.undertest.is_done(TEST_FILE_2), equal_to(True))
        assert_that(self.undertest.is_done(TEST_FILE_3), equal_to(True))

    def test_get_done(self):
        assert_that(self.undertest.get_done(".cands.astrom"), contains())
        assert_that(self.undertest.get_done(".reals.astrom"), contains())

        self.undertest.lock(TEST_FILE_1)
        self.undertest.record_done(TEST_FILE_1)
        self.undertest.unlock(TEST_FILE_1)
        assert_that(self.undertest.get_done(".cands.astrom"), contains(TEST_FILE_1))
        assert_that(self.undertest.get_done(".reals.astrom"), contains())

        self.undertest.lock(TEST_FILE_2)
        self.undertest.record_done(TEST_FILE_2)
        self.undertest.unlock(TEST_FILE_2)
        assert_that(self.undertest.get_done(".cands.astrom"), contains(TEST_FILE_1))
        assert_that(self.undertest.get_done(".reals.astrom"), contains(TEST_FILE_2))

        self.undertest.lock(TEST_FILE_3)
        self.undertest.record_done(TEST_FILE_3)
        self.undertest.unlock(TEST_FILE_3)
        assert_that(self.undertest.get_done(".cands.astrom"),
                    contains_inanyorder(TEST_FILE_1, TEST_FILE_3))
        assert_that(self.undertest.get_done(".reals.astrom"), contains(TEST_FILE_2))

    def test_lock_file(self):
        lock_requesting_user = "lock_requesting_user"

        self.undertest.lock(TEST_FILE_1)

        # No-one else should be able to acquire the lock...
        manager2 = self.create_independent_manager(userid=lock_requesting_user)
        self.assertRaises(FileLockedException, manager2.lock, TEST_FILE_1)

        # ... until we unlock it
        self.undertest.unlock(TEST_FILE_1)
        manager2.lock(TEST_FILE_1)

    def test_lock_holder_no_file_locked_exception(self):
        lock_holding_user = self.main_user_id
        lock_requesting_user = "lock_requesting_user"

        self.undertest.lock(TEST_FILE_1)

        # No-one else should be able to acquire the lock...
        manager2 = self.create_independent_manager(userid=lock_requesting_user)
        self.assertRaises(FileLockedException, manager2.lock, TEST_FILE_1)

        # ... but we should be able to without getting a FileLockedException
        self.undertest.lock(TEST_FILE_1)

    def test_lock_has_locker_id(self):
        lock_holding_user = self.main_user_id
        lock_requesting_user = "lock_requesting_user"

        file1 = TEST_FILE_1
        self.undertest.lock(file1)

        manager2 = self.create_independent_manager(userid=lock_requesting_user)

        try:
            manager2.lock(file1)
            self.fail("Should have thrown FileLockedExcecption")
        except FileLockedException as ex:
            assert_that(ex.filename, equal_to(file1))
            assert_that(ex.locker, equal_to(lock_holding_user))

    def test_record_done_puts_username_in_property(self):
        filename = TEST_FILE_1
        self.undertest.lock(filename)
        self.undertest.record_done(filename)
        self.undertest.unlock(filename)

        assert_that(self.get_property(filename, progress.DONE_PROPERTY),
                    equal_to(self.main_user_id))

    def test_unlock_after_record_done_no_error(self):
        file1 = TEST_FILE_1
        self.undertest.lock(file1)
        self.undertest.record_done(file1)
        self.undertest.unlock(file1)

    def test_record_done_does_not_unlock_all(self):
        file1 = TEST_FILE_1
        file2 = TEST_FILE_3
        manager2 = VOSpaceProgressManager(self.context)

        self.undertest.lock(file1)
        manager2.lock(file2)

        self.undertest.record_done(file1)
        assert_that(manager2.owns_lock(file2), equal_to(True))

        manager2.unlock(file2)
        assert_that(manager2.owns_lock(file2), equal_to(False))


class PartialProgressTrackingTest(AbstractVOSpaceProgressManagerTestCase, FileReadingTestCase):
    def _tracks_partial_progress(self):
        return True

    def test_record_index_requires_lock(self):
        self.assertRaises(RequiresLockException,
                          self.undertest.record_index,
                          TEST_FILE_1, 0)

    def test_record_index(self):
        self.undertest.lock(TEST_FILE_1)
        self.undertest.record_index(TEST_FILE_1, 1)
        self.undertest.record_index(TEST_FILE_1, 3)
        self.undertest.record_index(TEST_FILE_1, 0)

        assert_that(self.undertest.get_processed_indices(TEST_FILE_1),
                    contains_inanyorder(1, 3, 0))

        # Check they are still recorded after we release lock
        self.undertest.unlock(TEST_FILE_1)
        assert_that(self.undertest.get_processed_indices(TEST_FILE_1),
                    contains_inanyorder(1, 3, 0))

        # Check other clients can read them
        manager2 = self.create_independent_manager()
        assert_that(manager2.get_processed_indices(TEST_FILE_1),
                    contains_inanyorder(1, 3, 0))

    def test_get_processed_indices_empty_should_not_cause_error(self):
        assert_that(self.undertest.get_processed_indices(TEST_FILE_1),
                    has_length(0))

    def test_get_processed_indices_after_done(self):
        filename = TEST_FILE_1
        self.undertest.lock(filename)
        self.undertest.record_index(filename, 0)
        self.undertest.record_index(filename, 1)
        self.undertest.record_index(filename, 2)
        self.undertest.unlock(filename)

        assert_that(self.undertest.get_processed_indices(filename),
                    contains_inanyorder(0, 1, 2))

        self.undertest.lock(filename)
        self.undertest.record_done(filename)
        self.undertest.unlock(filename)

        assert_that(self.undertest.get_processed_indices(filename),
                    contains_inanyorder(0, 1, 2))

        # Double check with a second manager
        manager2 = self.create_independent_manager()
        assert_that(manager2.get_processed_indices(filename),
                    contains_inanyorder(0, 1, 2))


class NoTrackingTest(AbstractVOSpaceProgressManagerTestCase, FileReadingTestCase):
    def _tracks_partial_progress(self):
        return False

    def test_record_index_does_nothing(self):
        self.undertest.lock(TEST_FILE_1)
        self.undertest.record_index(TEST_FILE_1, 1)
        self.undertest.record_index(TEST_FILE_1, 3)
        self.undertest.record_index(TEST_FILE_1, 0)
        self.undertest.unlock(TEST_FILE_1)

        assert_that(self.has_property(TEST_FILE_1, progress.PROCESSED_INDICES_PROPERTY),
                    equal_to(False))

    def test_get_processed_indices_empty(self):
        self.undertest.lock(TEST_FILE_1)
        self.undertest.record_index(TEST_FILE_1, 1)
        self.undertest.record_index(TEST_FILE_1, 3)
        self.undertest.record_index(TEST_FILE_1, 0)
        self.undertest.unlock(TEST_FILE_1)

        assert_that(self.undertest.get_processed_indices(TEST_FILE_1), has_length(0))


if __name__ == '__main__':
    unittest.main()



__author__ = 'bannisterm'

'''
    Provide the LaTeX header and footer for a LaTeX AAS-journal style deluxtable that makes
    an OSSOS Data Release .detections file more human-readable.

    This is the format of table used in the OSSOS first-quarter paper.
'''



from contextlib import contextmanager

# no. columns must match no. c's
# table has small font and is horizontal.
header = r"\begin{deluxetable}{ccccccccccccc}" + '\n' + \
         r"\tabletypesize{\scriptsize}" + '\n' + \
         r"\rotate" + '\n' + \
         r"\tablecolumns{13}" + '\n' + \
         r"\tablehead{\colhead{$m_{r}$} \vspace{-0.2cm} & " \
         r"\colhead{$\sigma$ $m_{r}$} & " \
         r"\colhead{Detectability} & " \
         r"\colhead{RA} & " \
         r"\colhead{Dec} & " \
         r"\colhead{a} & " \
         r"\colhead{e} & " \
         r"\colhead{i} & " \
         r"\colhead{r$_{H}$} & " \
         r"\colhead{H$_{r}$} & " \
         r"\colhead{MPC} & " \
         r"\colhead{Object} & " \
         r"\colhead{Status} \\" \
         r"\colhead{discovery} & " \
         r"\colhead{all obs} & " \
         r"\colhead{} & " \
         r"\colhead{discov.} & " \
         r"\colhead{discov.} & " \
         r"\colhead{(AU)} & " \
         r"\colhead{} & " \
         r"\colhead{($^{\circ}$)} & " \
         r"\colhead{(AU)} & " \
         r"\colhead{} & " \
         r"\colhead{design.} & " \
         r"\colhead{} & " \
         r"\colhead{} " \
         + r"}" \
         + "\n" \
         + "\startdata \n" \
         + r"\cutinhead{Centaurs}" + "\n"

footer = r"\enddata " + "\n" + \
         r"\tablecomments{$p:q$: object is in the $p:q$ resonance; I: the orbit classification is currently " \
         r"insecure; " \
         r"" \
         r"" \
         r"H: the human operator intervened to declare the orbit security status. " \
         r"$a, e, i$ are J2000 ecliptic barycentric coordinates, with uncertainties from the covariant matrix fit " \
         r"of \citet{Bernstein:2000p444}; full barycentric elements are available at \url{http://www.ossos-survey" \
         r".org/}." \
         r"The full heliocentric orbital elements are available in electronic form from the Minor Planet Center.} " "\n" + \
         "\end{deluxetable} \n"


@contextmanager
def deluxe_table_formatter(outfile):
    ofile = open(outfile, 'w')
    ofile.write(header)

    yield ofile

    ofile.write(footer)
    ofile.close()


from __future__ import absolute_import

__author__ = 'jjk'

from numpy import (array, where)
import os
# from ossos.parsers import read_smooth_fit

path = '/Users/bannisterm/Dropbox/Papers in progress/OSSOS/First_quarter/data/'


def square_fit(m, params):
    return where(m < 21., params[0],
                    (params[0] - params[1] * (m - 21.0) ** 2) / (1. + np.exp((m - params[2]) / params[3])))


def square_fit_discovery_mag(obj, discov_mag, discov_rate):
    pwd = path + 'efficiency_motion_rates'
    blocks = {'o3e': "13AE", 'o3o': "13AO"}
    block = blocks[obj.lstrip('u')[0:3]]
    smooth_parameter_files = filter(lambda name: name.startswith('smooth'), os.listdir('{}/{}/'.format(pwd, block)))
    smooth_parameter_files.sort(key=lambda x: float(x.split('-')[1]))
    rates = array([float(x.split('-')[2]) for x in smooth_parameter_files])
    rate, = where(discov_rate < rates)
    fn = smooth_parameter_files[rate[0]]
    # pd, ps, mag_limit = read_smooth_fit('{}/{}/{}'.format(pwd, block, fn))
    eta = square_fit(discov_mag, ps)
    # print obj, discov_mag, eta, discov_rate, block, fn

    return float(eta)





__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from mock import Mock

from ossos.astrom import SourceReading
from ossos.downloads.async import DownloadRequest
from ossos.downloads.async import DownloadThread
from ossos.downloads.cutouts import ImageCutoutDownloader
from ossos.downloads.cutouts.source import SourceCutout


class DownloadRequestTest(unittest.TestCase):
    def test_execute_downloads_cutout_and_calls_callback(self):
        reading = Mock(spec=SourceReading)
        callback = Mock()

        downloader = Mock(spec=ImageCutoutDownloader)
        cutout = Mock(spec=SourceCutout)
        downloader.download_cutout.return_value = cutout

        request = DownloadRequest(reading, focus=None, needs_apcor=False,
                                  callback=callback)

        request.execute(downloader)

        downloader.download_cutout.assert_called_once_with(reading,
                                                           focus=reading.source_point,
                                                           needs_apcor=False)

        callback.assert_called_once_with(cutout)


class DownloadThreadTest(unittest.TestCase):
    def test_do_download_executes_request(self):
        work_queue = Mock()
        downloader = Mock()
        errorhandler = Mock()

        thread = DownloadThread(work_queue, downloader, errorhandler)

        request = Mock(spec=DownloadRequest)
        thread.do_download(request)

        request.execute.assert_called_once_with(downloader)


if __name__ == '__main__':
    unittest.main()



__author__ = 'Michele Bannister   git:@mtbannister'

import astropy  # specifically written for v 0.2.4. Later versions have different coordinate.ra/de print methods.

from ossos import mpc
from ossos import storage
from ossos.orbfit import Orbfit

date = '2015-07-20T12:00:00'
# lightcurve_targets = '/Users/michele/Dropbox/Telescope proposals/Subaru ' \
#                      'proposal_2015A_lightcurves/lightcurve_targets.txt'
ossinpath = 'vos:OSSOS/dbaseclone/ast/'  # '/Users/michele/Dropbox/OSSOS/measure3/ossin/'
outfile = '/Users/michele/Desktop/{}.txt'.format(date)
# '/Users/michele/Dropbox/Telescope proposals/Subaru proposal/lc_pos_20150414.txt'

with open(outfile, 'w') as ofile:
    ofile.write('Target	RA (hrs)	DEC		m_r	delta RA (")	delta DEC (")	Time predicted\n')

# with open(lightcurve_targets, 'r') as infile:
#     lines = infile.readlines()
#     for line in lines:
#         obj, mag = line.split('\t')
for kbo_filename in storage.listdir(ossinpath):
    mpc_observations = mpc.MPCReader(ossinpath + kbo_filename).mpc_observations
    orbit = Orbfit(mpc_observations)
    orbit.predict(date=date)

    with open(outfile, 'a') as ofile:
        ofile.write("{:>10s} {:>10s} {:>10s} {:6.2f} {:6.2f} {:>10s}\n".format(
            orbit.name, orbit.coordinate.ra.format(unit=astropy.units.core.Unit("hour"), sep=':'),
            orbit.coordinate.dec.format(sep=':'),  # need to add mag back in
            orbit.dra, orbit.ddec, date))



"""
The cutouts.source module contains SourceCutout class that provides access to the section of an image containing
the source.  The SourceCutout provides RA/DEC -> X/Y and X/Y -> RA/DEC mapping as well as cutout X/Y -> full image X/Y
"""

import traceback
import tempfile
from astropy import units
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.units import Quantity
from astropy.time import Time
from astropy.table import Table
import numpy
from ... import util


from ...astrom import SourceReading, Observation
from ...gui import logger
from ... import storage
from ossos.gui import config
from downloader import Downloader, ApcorData

__author__ = "David Rusk <drusk@uvic.ca>"


class SourceCutout(object):
    """
    A cutout around a source.
    """

    def __init__(self, reading, hdulist, radius=None):
        """
        :param reading: A source reading giving the measurement of the object associated with this cutout.
        :param hdulist: the HDUList containing the cutout.
        :return:
        """
        logger.debug("building a SourceCutout.")
        assert isinstance(reading, SourceReading)
        assert isinstance(hdulist, fits.HDUList)

        self.radius = radius
        self.reading = reading
        self.hdulist = hdulist
        self._adjusted = False
        self._apcor = None
        self._zmag = None
        self.init_skycoord = reading.sky_coord

        if self.reading.x is None or self.reading.y is None or (self.reading.x == -9999 and self.reading.y == -9999):
            # indicates that reading assocated with this SourceCutout doesn't have x/y/extension information.
            # Here we derive that information from the RA/DEC and the cutout parameters.
            (x, y, extno) = self.world2pix(self.reading.ra, self.reading.dec)
            self.reading.pix_coord = self.get_observation_coordinates(x, y, extno)
            self.reading.obs.ccdnum = self.get_ccdnum(extno)

        self._comparison_image = []
        self._comparison_image_index = None
        self._comparison_image_list = None
        self._tempfile = None
        self._bad_comparison_images = [self.hdulist[-1].header.get('EXPNUM', None)]

    def reset_coord(self):
        """
        Reset the source location based on the init_skycoord values
        @return:
        """
        (x, y, idx) = self.world2pix(self.init_skycoord.ra,
                                     self.init_skycoord.dec,
                                     usepv=True)
        self.update_pixel_location((x, y), idx)

    @property
    def pixel_coord(self):
        """
        Return the coordinates of the source in the cutout reference frame.
        @return:
        """
        return self.get_pixel_coordinates(self.reading.pix_coord, self.reading.get_ccd_num())

    @property
    def pixel_x(self):
        """
        The X coordinate of the reading in the cutout reference frame.
        @return: int
        """
        return self.pixel_coord[0]

    @property
    def pixel_y(self):
        """
        The Y coordiante of the reading in the cutout reference frame.
        @return:
        """
        return self.pixel_coord[1]

    @property
    def extno(self):
        """
        Which extension of the cutout is the current reading located in.
        A cutout can have multiple extensions.  The reading gives an x/y which refers to a
        specific extension.   If extno is not set then we can resolve the correct one.

        @return: int
        """

        return self.get_hdulist_idx(self.reading.get_ccd_num())

    def get_ccdnum(self, hdulist_index):
        """
        Get the CCDNUM for a given hdulist idx.

        @param hdulist_index:  the index of the HDUList entry that the CCDNUM is needed for.
        @return: ccdnum
        """
        return self.hdulist[hdulist_index].header.get('EXTVER')

    def get_hdulist_idx(self, ccdnum):
        """
        The SourceCutout is a list of HDUs, this method returns the index of the HDU that corresponds to the given
        ccd number.  CCDs are numbers from 0, but the first CCD (CCDNUM=0) is often in extension 1 of an MEF.
        @param ccdnum: the number of the CCD in the MEF that is being referenced.
        @return: the index of in self.hdulist that corresponds to the given CCD number.
        """
        for (extno, hdu) in enumerate(self.hdulist):
            if ccdnum == int(hdu.header.get('EXTVER', -1)) or str(ccdnum) in hdu.header.get('AMPNAME', ''):
                return extno
        raise ValueError("Failed to find requested CCD Number {} in cutout {}".format(ccdnum,
                                                                                      self))

    def flip_flip(self, point):
        """
        Sometimes we 'flipped' the X/Y coordinates of the image during processing.  This function takes an
        (x,y) location from the processing pipeline and returns the x/y location relative to the original
        reference frame.   (Correctly checks if the coordinate is flipped to begin with.)

        This is only for use on the original X/Y coordinates as provided by an input reading.

        @param point: an x/y location on the image.
        @type point: [x, y]
        @return:  the x/y location on the un-fliped frame
        """
        x, y = point
        if self.reading.compute_inverted():
            naxis1 = self.reading.obs.header.get(self.reading.obs.HEADER_IMG_SIZE_X,
                                                 self.reading.obs.header.get('NAXIS1', 256))
            naxis2 = self.reading.obs.header.get(self.reading.obs.HEADER_IMG_SIZE_Y,
                                                 self.reading.obs.header.get('NAXIS2', 256))
            logger.debug("Got that image has size {}, {}".format(naxis1, naxis2))

            x1 = x
            y1 = y
            x = int(naxis1) - x + 1
            y = int(naxis2) - y + 1
            logger.debug("Inverted source location from {},{} to {},{}".format(x1, y1, x, y))

        return x, y

    @property
    def astrom_header(self):
        return self.hdulist[len(self.hdulist) - 1].header

    @property
    def fits_header(self):
        return self.hdulist[len(self.hdulist) - 1].header

    @property
    def observed_source_point(self):
        return self.reading.pix_coord

    @property
    def pixel_source_point(self):
        return self.pixel_coord

    def update_pixel_location(self, new_pixel_location, hdu_index):
        """
        Update the x/y location of the associated reading by using new_pixel_location and transforming from cutout
        reference frame to observation refrence frame.
        @param new_pixel_location: (x, y) location of the source in cutout pixel reference frame
        @param hdu_index: index of the associate hdu in the hdulist for this cutout
        """
        self.reading.pix_coord = self.get_observation_coordinates(new_pixel_location[0],
                                                                  new_pixel_location[1],
                                                                  hdu_index)
        new_ccd = self.get_ccdnum(hdu_index)
        if new_ccd != self.reading.get_ccd_num():
            self._apcor = None
            self._zmag = None
            self.reading.obs.ccdnum = self.get_ccdnum(hdu_index)
        self.reading.sky_coord = self.pix2world(new_pixel_location[0],
                                                new_pixel_location[1],
                                                hdu_index)

    @property
    def ra(self):
        """
        Return the RA source.
        @return:
        """
        return self.reading.ra

    @property
    def dec(self):
        return self.reading.dec

    def is_adjusted(self):
        return self._adjusted

    def get_pixel_coordinates(self, point, ccdnum):
        """
        Retrieves the pixel location of a point within the current HDUList given the
        location in the original FITS image.  This takes into account that
        the image may be a cutout of a larger original.

        Args:
          point: tuple(float, float)
            (x, y) in original.

        Returns:
          (x, y) pixel in this image.
          @param extno: the extno from the original Mosaic that the x/y coordinates are from.
        """
        hdulist_index = self.get_hdulist_idx(ccdnum)
        if isinstance(point[0], Quantity) and isinstance(point[1], Quantity):
            pix_point = point[0].value, point[1].value
        else:
            pix_point = point
        if self.reading.inverted:
            pix_point = self.reading.obs.naxis1 - pix_point[0] +1 , self.reading.obs.naxis2 - pix_point[1] + 1

        (x, y) = self.hdulist[hdulist_index].converter.convert(pix_point)
        return x, y, hdulist_index

    def get_observation_coordinates(self, x, y, hdulist_index):
        """
        Retrieves the location of a point using the coordinate system of
        the original observation, i.e. the original image before any
        cutouts were done.

        Returns:
          (x, y) in the original image coordinate system.
          @param x: x-pixel location in the cutout frame of reference
          @param y: y-pixel location in the cutout frame of reference
          @param idx: index of hdu in hdulist that the given x/y corresponds to.
        """
        return self.hdulist[hdulist_index].converter.get_inverse_converter().convert((x, y))

    def pix2world(self, x, y, hdulist_index, usepv=True):
        ra, dec = self.hdulist[hdulist_index].wcs.xy2sky([x], [y], usepv=usepv)
        return ra[0], dec[0]

    def world2pix(self, ra, dec, usepv=True):
        """
        Convert a given RA/DEC position to the  X/Y/HDULIST_INDEX in the cutout frame.

         Here we loop over the hdulist to find the extension is RA/DEC are from and then return the appropriate X/Y
        :param ra: The Right Ascension of the point
        :type ra: Quantity
        :param dec: The Declination of the point
        :type dec: Quantity
        :return: X, Y, Extension position of source in cutout reference frame
        :rtype: (float, float, int)
        @param usepv: Should the non-linear part of the WCS be using to compute the transformation? (default: True)
        """
        x, y, idx = (0, 0, 0)
        for idx in range(1, len(self.hdulist)):
            hdu = self.hdulist[idx]
            x, y = hdu.wcs.sky2xy(ra, dec, usepv=usepv)
            if 0 < x < hdu.header['NAXIS1'] and 0 < y < hdu.header['NAXIS2']:
                # print "Inside the frame."
                return x, y, idx
        return x, y, idx

    @property
    def zmag(self,):
        """
        Return the photometric zeropoint of the CCD associated with the reading.
        @return: float
        """
        if self._zmag is None:
            hdulist_index = self.get_hdulist_idx(self.reading.get_ccd_num())
            self._zmag = self.hdulist[hdulist_index].header.get('PHOTZP', 0.0)
        return self._zmag

    @property
    def apcor(self):
        """
        return the aperture correction of for the CCD assocated with the reading.
        @return: Apcor
        """
        if self._apcor is None:
            try:
                self._apcor = Downloader().download_apcor(self.reading.get_apcor_uri())
            except:
                self._apcor = ApcorData.from_string("5 15 99.99 99.99")
        return self._apcor

    def get_observed_magnitude(self):
        # NOTE: this import is only here so that we don't load up IRAF
        # unnecessarily (ex: for candidates processing).
        """
        Get the magnitude at the current pixel x/y location.

        :return: Table
        """

        max_count = float(self.astrom_header.get("MAXCOUNT", 30000))
        (x, y, hdulist_index) = self.pixel_coord
        tmp_file = self._hdu_on_disk(hdulist_index)
        try:
            from ossos import daophot
            phot = daophot.phot_mag(tmp_file,
                                    x, y,
                                    aperture=self.apcor.aperture,
                                    sky=self.apcor.sky,
                                    swidth=self.apcor.swidth,
                                    apcor=self.apcor.apcor,
                                    zmag=self.zmag,
                                    maxcount=max_count, extno=1)
            if not self.apcor.valid:
                phot['PIER'][0] = 1
            return phot
        finally:
            self.close()

    def _hdu_on_disk(self, hdulist_index):
        """
        IRAF routines such as daophot need input on disk.

        Returns:
          filename: str
            The name of the file containing the FITS data.
        """
        if self._tempfile is None:
            self._tempfile = tempfile.NamedTemporaryFile(
                mode="r+b", suffix=".fits")
            self.hdulist[hdulist_index].writeto(self._tempfile.name)
        return self._tempfile.name

    def close(self):
        """
        Once we are done with the on disk content we should close the filehandle.
        """
        if self._tempfile is not None:
            self._tempfile.close()
            self._tempfile = None

    @property
    def comparison_image(self):
        if self.comparison_image_list[self.comparison_image_index]["REFERENCE"] is None:
            self.retrieve_comparison_image()
        row = self.comparison_image_list[self.comparison_image_list["ID"] == self.comparison_image_index]
        return row["REFERENCE"][0]

    @property
    def comparison_image_index(self):
        if self._comparison_image_index is None:
            ## Display a list of possible comparison images and ask user to select one.
            self.comparison_image_list.pprint()
            self.comparison_image_index = int(raw_input("SELECT ROW NUMBER OF DESIRED COMPARISON IMAGE: "))
        return self._comparison_image_index

    @comparison_image_index.setter
    def comparison_image_index(self, comparison_image_index):
        self._comparison_image_index = comparison_image_index

    @property
    def comparison_image_list(self):
        """
        returns a list of possible comparison images for the current cutout.  Will query CADC to create the list when
        first called.
        @rtype: Table
        """

        if self._comparison_image_list is not None:
            return self._comparison_image_list

        ref_ra = self.reading.ra * units.degree
        ref_dec = self.reading.dec * units.degree
        radius = self.radius is not None and self.radius or config.read('CUTOUTS.SINGLETS.RADIUS') * units.arcminute
        print "Querying CADC for list of possible comparison images at RA: {}, DEC: {}, raidus: {}".format(ref_ra,
                                                                                                           ref_dec,
                                                                                                           radius)
        query_result = storage.cone_search(ref_ra, ref_dec, radius, radius)  # returns an astropy.table.table.Table
        print "Got {} possible images".format(len(query_result))
        print "Building table for presentation and selection, including getting fwhm which is a bit slow."
        comparison_image_list = []
        if len(query_result['collectionID']) > 0:  # are there any comparison images even available on that sky?
            index = -1
            for row in query_result:
                expnum = row['collectionID']
                if expnum in self._bad_comparison_images:
                    continue
                date = Time(row['mjdate'], format='mjd').mpc
                exptime = row['exptime']
                filter_name = row['filter']
                if filter_name in self.hdulist[-1].header['FILTER']:
                    filter_name = "* {:8s}".format(filter_name)
                try:
                    fwhm = "{:5.2f}".format(float(storage.get_fwhm_tag(expnum, 22)))
                except:
                    fwhm = -1.00
                index += 1
                comparison_image_list.append([index, expnum, date, exptime, filter_name, fwhm, None])
        self._comparison_image_list = Table(data=numpy.array(comparison_image_list),
                                            names=["ID", "EXPNUM", "DATE-OBS", "FILTER", "EXPTIME", "FWHM", "REFERENCE"])
        return self._comparison_image_list


    def retrieve_comparison_image(self):
        """
        Search the DB for a comparison image for this cutout.
        """
        # selecting comparator when on a comparator should load a new one.
        collectionID = self.comparison_image_list[self.comparison_image_index]['EXPNUM']
        ref_ra = self.reading.ra * units.degree
        ref_dec = self.reading.dec * units.degree
        radius = self.radius is not None and self.radius or config.read('CUTOUTS.SINGLETS.RADIUS') * units.arcminute
        try:
            comparison = collectionID
            hdu_list = storage.ra_dec_cutout(storage.dbimages_uri(comparison),
                                             SkyCoord(ref_ra, ref_dec), radius)
            ccd = int(hdu_list[-1].header.get('EXTVER', 0))
            obs = Observation(str(comparison), 'p',
                              ccdnum=ccd)
            x = hdu_list[-1].header.get('NAXIS1', 0) // 2.0
            y = hdu_list[-1].header.get('NAXIS2', 0) // 2.0
            ref_ra = self.reading.ra * units.degree
            ref_dec = self.reading.dec * units.degree
            reading = SourceReading(x, y, self.reading.x, self.reading.y,
                                    ref_ra, ref_dec, self.reading.x, self.reading.y, obs)
            self.comparison_image_list[self.comparison_image_index]["REFERENCE"] = SourceCutout(reading, hdu_list)
        except Exception as ex:
            print traceback.format_exc()
            print ex
            print "Failed to load comparison image;"
            self.comparison_image_index = None
            logger.error("{} {}".format(type(ex), str(ex)))
            logger.error(traceback.format_exc())







__author__ = "David Rusk <drusk@uvic.ca>"

import tempfile
import unittest

from hamcrest import assert_that, equal_to, contains
from mock import Mock, call, MagicMock

import vos

from tests.base_tests import FileReadingTestCase
from ossos.astrom import SourceReading
from ossos.downloads.cutouts import ImageCutoutDownloader


class CutoutDownloaderTest(FileReadingTestCase):
    def setUp(self):
        self.image_uri = "vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1584431/1584431p15.fits"
        self.apcor_uri = "vos://cadc.nrc.ca~vospace/OSSOS/dbimages/1584431/ccd15/1584431p15.apcor"

        self.reading = MagicMock(spec=SourceReading)
        self.reading.x = 0
        self.reading.y = 0
        self.reading.ra = 0
        self.reading.dec = 0
        self.reading.get_image_uri.return_value = self.image_uri
        self.reading.get_apcor_uri.return_value = self.apcor_uri
        self.reading.get_extension.return_value = 19
        self.reading.get_original_image_size.return_value = (2000, 3000)
        self.reading.is_inverted.return_value = False

        self.needs_apcor = True
        self.focus = (75, 80)

        self.vosclient = Mock(spec=vos.Client)
        self.downloader = ImageCutoutDownloader(slice_rows=100, slice_cols=50,
                                                vosclient=self.vosclient)

        # Mock vosclient to open a local file instead of one from vospace
        self.localfile = open(self.get_abs_path("data/testimg.fits"), "rb")
        self.apcorfile = tempfile.TemporaryFile("r+b")
        self.apcorfile.write("4 10 0.3 0.1")
        self.apcorfile.flush()
        self.apcorfile.seek(0)

        def choose_ret_val(*args, **kwargs):
            selected_file = None
            if self.image_uri in args:
                selected_file = self.localfile
            elif self.apcor_uri in args:
                selected_file = self.apcorfile
            else:
                self.fail("Unrecognized URI")

            selected_file.seek(0)

            return selected_file

        self.vosclient.open.side_effect = choose_ret_val

    def tearDown(self):
        self.localfile.close()
        self.apcorfile.close()

    def test_request_image_cutout(self):
        download = self.downloader.download_cutout(self.reading,
                                                   focus=self.focus,
                                                   needs_apcor=self.needs_apcor)

        assert_that(self.vosclient.open.call_args_list, contains(
            call(self.image_uri, view="cutout", cutout="[19][50:100,30:130]"),
            call(self.apcor_uri, view="data")
        ))

        # This is just a test file, make sure we can read an expected value
        # it.  It won't have the right shape necessarily though.
        assert_that(download.fits_header["FILENAME"],
                    equal_to("u5780205r_cvt.c0h"))


if __name__ == '__main__':
    unittest.main()



__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from mock import Mock, MagicMock

from tests.base_tests import WxWidgetTestCase
from tests.testutil import mock_hdulist
from ossos.astrom import SourceReading
from ossos.downloads.cutouts.source import SourceCutout
from ossos.fitsviewer.singletviewer import SingletViewer


class SingletViewerTest(WxWidgetTestCase):
    def setUp(self):
        super(SingletViewerTest, self).setUp()

        self.canvas = Mock()
        self.viewer = SingletViewer(self.rootframe, self.canvas)
        self.do_render = Mock()
        self.viewer._do_render = self.do_render

    def test_mark_source(self):
        x, y = 50, 100
        fwhm = 4

        cutout = Mock(spec=SourceCutout)
        reading = Mock(spec=SourceReading)
        reading.null_observation = Mock(return_value=False)
        cutout.hdulist = mock_hdulist()
        cutout.pixel_source_point = x, y
        cutout.astrom_header = {"FWHM": fwhm}
        reading.from_input_file = False
        cutout.reading = reading

        self.viewer.display(cutout, mark_source=False)

        current_displayable = self.viewer.current_displayable
        mock_place_marker = Mock()
        current_displayable.place_marker = mock_place_marker

        self.viewer.mark_sources(cutout)
        mock_place_marker.assert_called_once_with(x, y, 2 * fwhm,
                                                  colour="b")

    def test_refresh_marker(self):
        cutout = Mock(spec=SourceCutout)
        cutout.hdulist = mock_hdulist()

        mark_source = Mock()
        self.viewer.mark_sources = mark_source

        self.viewer.display(cutout, mark_source=False)

        self.viewer.refresh_markers()
        mark_source.assert_called_once_with(cutout)

    def test_toggle_reticule(self):
        cutout = Mock(spec=SourceCutout)
        cutout.hdulist = mock_hdulist()

        mark_source = Mock()
        self.viewer.mark_sources = mark_source

        self.viewer.display(cutout)

        toggle_reticule = Mock()
        current_displayable = self.viewer.current_displayable
        current_displayable.toggle_reticule = toggle_reticule

        self.viewer.toggle_reticule()

        toggle_reticule.assert_called_once_with()


if __name__ == '__main__':
    unittest.main()
